# Copyright (c) Microsoft Corporation.  All rights reserved.

"""
This package contains the PROSE code acceleration SDK. Code Accelerator
uses the power of PROSE to quickly and accurately produce Python code for
data preparation tasks--making you faster and more productive.
"""
from ._api import (
    # APIs
    derive_value,
    _derive_value,
    detect_types,
    read,
    read_html,
    _read_html,
    read_json,
    read_parquet,
    read_text,
    _read_text,
    # classes
    DeriveValueOptions,
    DeriveValueResult,
    DetectTypesOptions,
    DetectTypesResult,
    DomainTypeProfileBuilder,
    DomainTypeProfileResult,
    FindPatternsBuilder,
    FindPatternsLearnResult,
    ReadHtmlOptions,
    ReadHtmlResult,
    ReadJsonOptions,
    ReadJsonResult,
    ReadParquetOptions,
    ReadParquetResult,
    ReadTextOptions,
    ReadTextResult,
)

from ._config import config
from ._version import __version__

__all__ = [
    # APIs
    "config",
    "detect_types",
    "read",
    "read_html",
    "read_json",
    "read_parquet",
    "read_text",
    "derive_value",
    # classes
    "DeriveValueOptions",
    "DeriveValueResult",
    "DetectTypesOptions",
    "DetectTypesResult",
    "DomainTypeProfileBuilder",
    "DomainTypeProfileResult",
    "FindPatternsBuilder",
    "FindPatternsLearnResult",
    "ReadHtmlOptions",
    "ReadHtmlResult",
    "ReadJsonOptions",
    "ReadJsonResult",
    "ReadParquetOptions",
    "ReadParquetResult",
    "ReadTextOptions",
    "ReadTextResult",
]


def _jupyter_nbextension_paths():
    return [
        {
            "section": "notebook",
            "src": "../static",
            "dest": "prose",
            "require": "prose/extension",
        }
    ]


_TELEMETRY_MESSAGE = """
    PROSE CodeAccelerator SDK Telemetry
    -----------------------------------
    The PROSE CodeAccelerator SDK includes a telemetry feature that collects
    usage information. It is important that the PROSE team understands how the
    SDK is being used so that we can improve it.

    The data collected is not used to identify any person.

    The PROSE CodeAccelerator SDK telemetry feature is enabled by default. You
    can opt-out of the telemetry feature by setting an environment variable
    PROSE_TELEMETRY_OPTOUT (for example, 'export' on macOS/Linux, 'set' on
    Windows) to any value.
"""


# The filename whose existence flags that the telemetry message has been displayed to a user
def _telemetry_message_filename():
    import os

    return os.path.join(
        os.path.expanduser("~"), ".config", "prose", "telemetry_message"
    )


def _display_telemetry_message():
    """Display telemetry message if not already displayed."""
    import errno, os

    # The mshost environment already provided consent information for telemetry.
    if os.environ.get("MSHOST_TELEMETRY_ENABLED", "").lower() in ("true", "false"):
        return

    # Display message if the user hasn't seen the telemetry message, based on
    # the existance of a telemetry message file.
    filename = _telemetry_message_filename()
    if not os.path.isfile(filename):
        telemetry_message_dir = os.path.dirname(filename)
        try:
            os.makedirs(telemetry_message_dir)
        except OSError as ex:
            if ex.errno == errno.EEXIST and os.path.isdir(telemetry_message_dir):
                pass
            else:
                raise
        with open(filename, "w"):
            pass
        print(_TELEMETRY_MESSAGE)


# Show telemetry message if necessary
_display_telemetry_message()
