# Copyright (c) Microsoft Corporation.  All rights reserved.

import numpy as np

from ._utils import (
    record_telemetry,
    record_telemetry_reuse_logger,
    _serialize_dataframe,
    _deserialize_dataframe,
)
from ._shared_utils import create_mimebundle
from .._derive_value_result import DeriveValueResultImpl
from ._derive_value_options import DeriveValueOptions


class DeriveValueResult(object):
    """The result from attempting to learn a program for deriving a value from input value(s).

    (The result of calling :meth:`prose.codeaccelerator.derive_value`.)
    """

    @record_telemetry_reuse_logger
    def __init__(self, logger, df, options):
        """Creates a result object for learning code to generate values for a new column.
        This should only be called internally by the library.
        """

        self._impl = DeriveValueResultImpl(logger, df, options)

    @property
    @record_telemetry(accessor="get")
    def code(self):
        """Return the code that was learned.

        :return: The string representation of synthesized code snippet.
        :rtype: str
        """
        return self._impl.code

    @property
    @record_telemetry(accessor="get")
    def options(self):
        """DeriveValueOptions object containing values that affect the generated code.

        :return:
        :rtype: :class:`prose.codeaccelerator.DeriveValueOptions`
        """
        return self._impl.options

    @property
    @record_telemetry(accessor="get")
    def data(self):
        """Return the new dataframe containing a new column obtained by running the synthesized program on the input dataframe.

        :return: The dataframe containing the new column.
        :rtype: `DataFrame <https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.html>`
        """
        return self._impl.data

    @record_telemetry
    def update(self):
        """(Re)learn a program and update the code based on the current options."""

        return self._impl.learn()

    def to_serializable(self):
        """Return a serializable representation of this explicitly for use by our notebook front end.
        This does not completely round trip an arbitrary result object--only one that was created by the front end.
        """
        if not self.options._index_based_examples and self.options.examples:
            raise ValueError(
                "Serialization of result object only supports index_based_examples (and not dict-like examples"
            )
        return {
            "result_type": "DeriveValue",
            "data_frame_json": _serialize_dataframe(self.data),
            "index_based_examples": self.options._index_based_examples or [],
            "input_column_names": self.options.input_column_names,
            "new_column_name": self.options.new_column_name,
            "bound_names": self.options._bound_names or [],
            "input_table_variable_name": self.options._input_table_variable_name,
            "code": self.code,
        }

    @classmethod
    def _from_serializable(cls, logger, serializable):
        """Construct an instance of this class cls using the df and options given in serializable form"""
        index_based_examples = serializable["index_based_examples"]
        new_column_name = serializable["new_column_name"]
        df = _deserialize_dataframe(serializable["data_frame_json"])
        # remove the column that was added by the last learn call
        if new_column_name in df.columns:
            del df[new_column_name]

        options = DeriveValueOptions(
            examples=None,
            input_column_names=serializable["input_column_names"],
            new_column_name=new_column_name,
            bound_names=serializable["bound_names"],
            input_table_variable_name=serializable["input_table_variable_name"],
            index_based_examples=serializable["index_based_examples"],
        )
        options._index_based_examples = index_based_examples

        return cls(logger, df, options)

    @record_telemetry
    def _repr_mimebundle_(self, include, exclude):
        """Return an object passed to the mime type renderer in the client side extensions."""
        return create_mimebundle(self)

    @record_telemetry
    def __repr__(self):
        """Return a printable string that summarizes the current state of the derive column process."""

        return self._impl.__repr__()
