# Copyright (c) Microsoft Corporation.  All rights reserved.

from prose.codeaccelerator import _internals, _utils
from ._read_options import ReadOptions
from .._config import config


class DetectTypesOptions(ReadOptions):
    """Options specific to type detection."""

    _supported_type_names = {
        "bool",
        "boolean",
        "numeric",
        "number",
        "datetime",
        "string",
        "text",
    }

    @property
    def _internal(self):
        return self._detect_types_options

    def __init__(
        self, target=None, types_to_detect=None, rows_to_analyze=None, true_length=None
    ):
        """Initializes options for type detection.

        :param target: Target platform for generated code.
            Supported values are :code:`"auto"`, :code:`"pandas"`,
            and :code:`"pyspark"`.
        :param types_to_detect: The list of types to detect.
            Valid strings in the list include the values: 'bool', 'boolean',
            'numeric', 'number', 'datetime', 'string' and 'text'.
        :param rows_to_analyze: Number of rows to analyze while detecting data types.
            The value :code:`None` means all rows are analyzed.
        :param true_length: The actual size of the dataset, in the case that
            the input data is a sample of the entire dataset.
        """
        self._detect_types_options = _internals.DetectTypesOptions()
        self.target = target if target else DetectTypesOptions._default_target()
        if types_to_detect is not None:
            self.types_to_detect = types_to_detect
        if rows_to_analyze is not None:
            self.rows_to_analyze = rows_to_analyze
        if true_length is not None:
            self.true_length = true_length

    @classmethod
    def _default_target(cls):
        """Default target to use for this option."""
        return config.get("default_target", "auto")

    def _allowed_targets(self):
        """Specify a list of the allowed targets for this builder type."""
        return self._detect_types_options.allowed_targets

    @property
    def target(self):
        """Target runtime library for the generated code.

        Supported values are :code:`"auto"`, :code:`"pandas"`, and :code:`"pyspark"`.

        :type: str
        """
        return self._detect_types_options.target

    @target.setter
    def target(self, value):
        lower_value = super()._validate_target(value)
        if lower_value == "pyspark" and self.mode == "list":
            raise ValueError(
                "Cannot set target to 'pyspark' when input data is list. "
                "Try passing a dict or pandas dataframe for pyspark code generation."
            )
        self._detect_types_options.target = lower_value

    @property
    def mode(self):
        """The mode of the generated code (based on the input data).
        Possible values: :code:`"list"`, :code:`"dictionary"`, :code:`"dataframe"`, :code:`"pyspark"`.
        """
        return self._detect_types_options.mode

    def _set_mode(self, value):
        """Set the code generation mode. For internal use only."""
        self._detect_types_options.mode = value

    @property
    def true_length(self):
        """The actual size of the dataset, in the case that the input data is
        a sample of the entire dataset.

        :type: int
        """
        return self._detect_types_options.true_length

    @true_length.setter
    def true_length(self, value):
        if (value is not None) and value < 1:
            raise ValueError("The true length must be at least 1.")
        self._detect_types_options.true_length = value

    @property
    def rows_to_analyze(self):
        """Number of rows to analyze while detecting data types.
        
        The value :code:`None`, which is the default, means all rows are analyzed.

        :type: int
        """
        return self._detect_types_options.rows_to_analyze

    @rows_to_analyze.setter
    def rows_to_analyze(self, value):
        if (value is not None) and value < 1:
            raise ValueError("The number of rows to analyze must be at least 1.")
        self._detect_types_options.rows_to_analyze = value

    @property
    def types_to_detect(self):
        """The list of types to detect.

        The value `None` which is the default means
        that the standard list of types, which includes Booleans, Numeric types,
        Dates and Times, and String valued types are to be detected.
        Valid strings in the list include the values: 'bool', 'boolean',
        'numeric', 'number', 'datetime', 'string' and 'text'.

        If the values in a column are not identifiable as values
        of the types that `types_to_detect` is set to, then they will
        be treated as string values.
        """
        return self._detect_types_options.types_to_detect

    @types_to_detect.setter
    def types_to_detect(self, value):
        if value is None:
            return

        new_value = list(value)
        if not all(isinstance(x, str) for x in new_value):
            raise _utils.ProseException(
                "types_to_detect can only be assigned a list of string values."
            )
        for t in new_value:
            if t not in self._supported_type_names:
                raise _utils.ProseException(
                    "Unknown type name '%s'. Supported type names are: %s"
                    % (t, ", ".join(sorted(self._supported_type_names)))
                )
        self._detect_types_options.types_to_detect = value

    @property
    def detected_types(self):
        """The list of the learned/detected types."""
        return self._detect_types_options.detected_types

    def __repr__(self):
        return """\
DetectTypesOptions
===============
target = {}
mode = {}
rows_to_analyze = {}
types_to_learn = {}
detected_types = {}
true_length = {}""".format(
            self.target,
            self.mode,
            self.rows_to_analyze,
            self.types_to_detect,
            self.detected_types,
            self.true_length,
        )
