# Copyright (c) Microsoft Corporation.  All rights reserved.

from prose.codeaccelerator._domain_type_profile_builder import (
    DomainTypeProfileBuilderImpl,
)
from ._utils import (
    record_telemetry_for_builder_api,
    record_derived_telemetry,
    record_new_telemetry,
)
from io import StringIO


class DomainTypeProfileBuilder(object):
    """A builder that can be used to produce profiles of a column of data or the columns of a dataframe,
    wrt a set of a domain types.

    :param data: The data to profile, must be one of (a) a string-valued list-like object, or
        (b) a pandas.DataFrame with string-valued columns."""

    @record_new_telemetry
    def __init__(self, data):
        """Create a DomainTypeProfileBuilder to profile data with respect to a set of domain types."""
        self._impl = None

        def impl(logger):
            self._impl = DomainTypeProfileBuilderImpl(logger, data)

        # noinspection PyReturnFromInit
        return impl

    @property
    @record_telemetry_for_builder_api(accessor="get")
    def target(self):
        """This is unused and ignored."""
        return self._impl.target

    @target.setter
    @record_telemetry_for_builder_api(accessor="set")
    def target(self, value):
        self._impl.target = value

    @property
    @record_telemetry_for_builder_api(accessor="get")
    def types_to_detect(self):
        """The list of types to detect."""
        return self._impl.types_to_detect

    @types_to_detect.setter
    @record_telemetry_for_builder_api(accessor="set")
    def types_to_detect(self, value):
        self._impl.types_to_detect = value

    @record_derived_telemetry("DomainTypeProfileResult")
    def learn(self):
        """Infer the domain types in the provided data.

        :return: A DomainTypeProfileResult object.
        """
        return lambda logger: DomainTypeProfileResult(self._impl.learn(logger))

    @record_telemetry_for_builder_api
    def __repr__(self):
        return self._impl.__repr__()


class DomainTypeProfileResult(object):
    """Represents the result produced by a call to learn() on a DomainTypeProfileBuilder object."""

    def __init__(self, result):
        self._impl = result

    def __getitem__(self, item) -> "SingleColumnProfileResult":
        """Returns the profile information for a single column from the result.

        :param item: The name of the column for which the profile result is to be obtained."""
        return SingleColumnProfileResult(self._impl[item])

    def keys(self):
        return self._impl.keys()

    def __repr__(self):
        with StringIO() as b:
            b.write("DomainTypeProfileResult {\n")
            first = True
            for key in self.keys():
                if not first:
                    b.write("\n")
                first = False

                current = repr(self[key]).splitlines()
                for line in current:
                    b.write("    " + line + "\n")

            b.write("}\n")
            return b.getvalue()


class SingleColumnProfileResult(object):
    def __init__(self, result):
        self._impl = result

    @property
    def target_column(self):
        """The column corresponding to this profile result"""
        return self._impl.target_column

    @property
    def dominant_domain_type_name(self):
        """The name of the domain type that dominates values in this column."""
        return self._impl.dominant_domain_type_name

    @property
    def dominant_domain_type_count(self):
        """The count of the number of values belonging to the dominant domain type in the column."""
        return self._impl.dominant_domain_type_count

    @property
    def null_string_count(self):
        """A count of the number of null strings in the column."""
        return self._impl.null_string_count

    @property
    def empty_string_count(self):
        """A count of the number of empty strings in the column."""
        return self._impl.empty_string_count

    @property
    def sample_count(self):
        """A count of the number of strings in the column."""
        return self._impl.sample_count

    @property
    def column_name(self):
        """The name of the column."""
        return self._impl.column_name

    def __repr__(self):
        with StringIO() as b:
            b.write("Column Name:    " + self.column_name + "\n")
            b.write("Domain Type:    " + self.dominant_domain_type_name + "\n")
            b.write(
                "Match Fraction: %f"
                % (self.dominant_domain_type_count / self.sample_count)
                + "\n"
            )
            if self.empty_string_count > 0:
                b.write(
                    "Empty Fraction: %f" % (self.empty_string_count / self.sample_count)
                    + "\n"
                )
            if self.null_string_count > 0:
                b.write(
                    "Null Fraction:  %f" % (self.null_string_count / self.sample_count)
                    + "\n"
                )

            return b.getvalue()
