# Copyright (c) Microsoft Corporation.  All rights reserved.

from .._find_patterns_builder import (
    FindPatternsBuilderImpl,
    FindPatternsLearnResultImpl,
)
from ._utils import (
    record_telemetry_for_builder_api,
    record_derived_telemetry,
    record_new_telemetry,
)


class FindPatternsBuilder:
    """A builder for data and constraints in order to learn code for identifying patterns in a set of strings.

    :param data: List of strings to identify patterns from.
    :type data: List[str]
    """

    @record_new_telemetry
    def __init__(self, data):
        def impl(logger):
            self._impl = FindPatternsBuilderImpl(logger, data)

        return impl

    @property
    @record_telemetry_for_builder_api(accessor="get")
    def target(self):
        """Target runtime library for the generated code.

        Supported values are `"auto"`, `"pandas"`, and `"pyspark"`. Default is `"pandas"`.
        """
        return self._impl.target

    @target.setter
    @record_telemetry_for_builder_api(accessor="set")
    def target(self, value):
        self._impl.target = value

    @record_derived_telemetry("FindPatternsLearnResult")
    def learn(self):
        """Attempt to learn a set of patterns that cover the data and satisfy the constraints that have been set on
        this builder. """
        return lambda logger: FindPatternsLearnResult(self._impl.learn(logger))

    @property
    @record_telemetry_for_builder_api(accessor="get")
    def in_same_cluster(self):
        """A list of collections of strings. Each collection of strings constrains
        all the strings in the collection to be covered by the same regex/cluster.
        Each collection should have at least 2 strings, and all strings in the
        collections must come from the data provided to the builder during construction.

        :rtype: List[Iter[str]]
        """
        return self._impl.in_same_cluster

    @in_same_cluster.setter
    @record_telemetry_for_builder_api(accessor="set")
    def in_same_cluster(self, value):
        self._impl.in_same_cluster = value

    @property
    @record_telemetry_for_builder_api(accessor="get")
    def in_different_clusters(self):
        """A list of collections of strings. Each collection of strings constrains
        each string in the collection to be covered by a different regex/cluster.
        Each collection should have at least 2 strings, and all strings in the
        collections must come from the data provided to the builder during construction.

        :rtype: List[Iter[str]]
        """
        return self._impl.in_different_clusters

    @in_different_clusters.setter
    @record_telemetry_for_builder_api(accessor="set")
    def in_different_clusters(self, value):
        self._impl.in_different_clusters = value

    @property
    @record_telemetry_for_builder_api(accessor="get")
    def custom_regex_tokens(self):
        """A list of custom regex tokens -- each custom regex token is represented as a tuple
        of strings (name, regex). The custom regex tokens are used while generating the patterns.

        :rtype: List[Tuple[str, str]]
        """
        return self._impl.custom_regex_tokens

    @custom_regex_tokens.setter
    @record_telemetry_for_builder_api(accessor="set")
    def custom_regex_tokens(self, value):
        self._impl.custom_regex_tokens = value

    @property
    @record_telemetry_for_builder_api(accessor="get")
    def optimize_for_long_constants(self):
        """A switch controlling whether the `.learn()` uses the long constant optimization.
        This optimization improves performance significantly for input data with long repeating constants, while
        potentially producing marginally worse patterns. The default behaviour (when the value is `None`) is to enable
        heuristically enable the optimization, based on the characteristics of the inputs. The optimization
        can be enabled or disabled in other cases by setting the value to `True` and `False`, respectively.

        :rtype: Optional[bool]
        """
        return self._impl.optimize_for_long_constants

    @optimize_for_long_constants.setter
    @record_telemetry_for_builder_api(accessor="set")
    def optimize_for_long_constants(self, value):
        self._impl.optimize_for_long_constants = value

    @property
    @record_telemetry_for_builder_api(accessor="get")
    def include_outlier_patterns(self):
        """A switch controlling whether the `.learn()` returns outlier patterns too. These outlier
        patterns may be of a lower quality and may cover significantly fewer inputs. By default, it is set to `False`.

        :rtype: bool
        """
        return self._impl.include_outlier_patterns

    @include_outlier_patterns.setter
    @record_telemetry_for_builder_api(accessor="set")
    def include_outlier_patterns(self, value):
        self._impl.include_outlier_patterns = value

    @record_telemetry_for_builder_api
    def __repr__(self):
        """Return a printable string that summarizes the data and the constraints in this builder."""
        return self._impl.__repr__()


class FindPatternsLearnResult:
    """The result from attempting to learn a set of patterns in string data."""

    def __init__(self, result):
        """Create a learning result from a find patterns operation. Only intended to be called internally.

        :param data: The data that was used for learning.
        :param result: The result that was returned from _internals.FindPatternsBuilder.learn.
        """
        self._impl = result

    @property
    @record_telemetry_for_builder_api(accessor="get")
    def regexes(self):
        """List of regexes representing the patterns identified in the data.

        `None` represents a pattern that only matches strings that are `None`.

        :rtype: List[str]
        """
        return self._impl.regexes

    @record_telemetry_for_builder_api
    def code(self, task="classify"):
        """Return a code fragment that identifies the patterns that were learned.

        The code is targeted to the platform specified by the corresponding `FindPatternsBuilder.target`.

        :param task: A parameter that can take the values 'check' or 'classify'.
            If the parameter is 'classify', the method produces a code fragment to identify
            the pattern that classifies strings. If the parameter is 'check', the method
            produces a code fragment to assert that a given string falls into one of the
            patterns learned.
        """
        return self._impl.code(task)

    @property
    @record_telemetry_for_builder_api(accessor="get")
    def preview_data(self):
        """Group data provided to the find patterns operation by the learned patterns.

        :returns: A dictionary with keys being the name of the pattern, and the values
            being a list of strings that correspond to that pattern.
        """
        return self._impl.preview_data

    @record_telemetry_for_builder_api
    def __repr__(self):
        """Return a printable string that summarizes the results of the learning operation."""
        return self._impl.__repr__()
