# Copyright (c) Microsoft Corporation.  All rights reserved.
from ._utils import record_new_telemetry_for_top_level
from .._buffered_reader import ReadTextBufferedReader
from ..exceptions import LearnError
from .._internals import DetectFileType
from ._read_parquet_result import ReadParquetResult
from ._read_json_result import ReadJsonResult
from ._read_text_result import ReadTextResult
from ._read_html_result import ReadHtmlResult


@record_new_telemetry_for_top_level
def read(filename):
    """
    Detects the type and generates appropriate code to read a file specified in
    :filename: into a table.

    :param filename: The file to read the table from.
    :type filename: str
    """

    def impl(logger):
        reader = ReadTextBufferedReader(filename, create_stream=True)
        dt = DetectFileType(logger, filename, reader)
        t = dt.detect()
        if t == "text":
            return ReadTextResult(logger, filename, None)
        elif t == "json":
            return ReadJsonResult(logger, filename, None)
        elif t == "parquet":
            return ReadParquetResult(logger, filename, None)
        else:
            raise LearnError("Could not detect file type.")

    return impl


@record_new_telemetry_for_top_level
def read_parquet(filename, options=None):
    """Generates code to read a parquet file specified in :filename: with given
    options in :options:

    :param filename: The parquet file to read the table from.
    :type filename: str
    :param options: Options to specify extra settings to learn code.
    :type options: :class `prose.codeaccelerator.ReadParquetOptions`

    :return: The result of learning.
    :rtype: :class:'prose.codeaccelerator.ReadParquetResult`.
    """
    return lambda logger: ReadParquetResult(logger, filename, options)


@record_new_telemetry_for_top_level
def read_text(filename, options=None):
    """Generates code to read a text file specified in :filename: with given
    options in :options:

    :param filename: The text file to read the table from.
    :type filename: str
    :param options: Options to specify extra settings to learn code.
    :type options: :class `prose.codeaccelerator.ReadTextOptions`

    :return: The result of learning.
    :rtype: :class:'prose.codeaccelerator.ReadTextResult`.
    """
    return lambda logger: ReadTextResult(logger, filename, options)


@record_new_telemetry_for_top_level
def _read_text(mime_data):
    """Internal version of the read_text entry point function which takes mime
    bundle data object, runs the learn call and returns an updated mime bundle
    data object. Importantly this is not the entire bundle, just the data
    portion.
    """

    def logged_read_text(logger):
        result = ReadTextResult.from_serializable(logger, mime_data)
        return result.to_serializable()

    return logged_read_text


@record_new_telemetry_for_top_level
def read_json(filename, options=None):
    """Generates code to read a JSON file specified in :filename: with given
    options in :options:

    :param filename: The JSON file to read the table from.
    :type filename: str
    :param options: Options to specify extra settings to learn code.
    :type options: :class `prose.codeaccelerator.ReadJsonOptions`

    :rtype: :class:'prose.codeaccelerator.ReadJsonResult`.
    """
    return lambda logger: ReadJsonResult(logger, filename, options)


@record_new_telemetry_for_top_level
def read_html(filename, options=None):
    """Generates code to read a HTML file specified in :filename: with given
    options in :options:

    :param filename: The HTML file to read the table from.
    :type filename: str
    :param options: Options to specify extra settings to learn code.
    :type options: :class `prose.codeaccelerator.ReadHtmlOptions`
    :rtype: :class:'prose.codeaccelerator.ReadHtmlResult`.
    """
    return lambda logger: ReadHtmlResult(logger, filename, options)


def _read_html(mime_data):
    """Internal version of the read_html entry point function which takes mime
    bundle data object, runs the learn call and returns an updated mime bundle
    data object. Importantly this is not the entire bundle, just the data
    portion.
    """
    result = ReadHtmlResult.from_serializable(mime_data)
    return result.to_serializable()
