# Copyright (c) Microsoft Corporation.  All rights reserved.

from prose.codeaccelerator import _internals
from ._read_options import ReadOptions
from .._buffered_reader import ReadTextBufferedReader
from .._config import config


class ReadJsonOptions(ReadOptions):
    """Options specific to reading JSON files."""

    @property
    def _internal(self):
        return self._read_json_options

    def __init__(self, target=None, lines_to_analyze=None, detect_data_types=False):
        """Initializes options for reading JSON file.

        :param target: Target platform for generated code.
        :param lines_to_analyze: Number of lines of the input file that should be used during learning.
        :param detect_data_types: Whether or not to synthesize code to detect data types as part of reading the file.
        """
        self._read_json_options = _internals.ReadJsonOptions()
        self.target = target if target else ReadJsonOptions._default_target()
        if lines_to_analyze:
            self.lines_to_analyze = lines_to_analyze
        self.detect_data_types = detect_data_types

    @classmethod
    def _default_target(cls):
        """Default target to use for this option."""
        return config.get("default_target", "pandas")

    def _allowed_targets(self):
        """Specify a list of the allowed targets for this builder type."""
        return self._read_json_options.allowed_targets

    @property
    def target(self):
        """Target runtime library for the generated code.

        Supported values are :code:`"pandas"` and :code:`"pyspark"`.

        :type: str
        """
        return self._read_json_options.target

    @target.setter
    def target(self, value):
        value = super()._validate_target(value)
        self._read_json_options.target = value

    @property
    def lines_to_analyze(self):
        """Number of lines of the input file that should be used during learning.

        :type: int
        """
        return self._read_json_options.lines_to_analyze

    @lines_to_analyze.setter
    def lines_to_analyze(self, value):
        if value < 1:
            raise ValueError("The number of lines to analyze must be at least 1.")
        self._read_json_options.lines_to_analyze = value

    @property
    def encoding(self):
        """The detected encoding of the sample file that will be used for subsequent reads.

        :type: str
        """
        return self._read_json_options.encoding

    @property
    def names(self):
        """List of column names learned from the input file that will be used in the output table.
        
        :type: List[str]
        """
        return self._read_json_options.names

    def __repr__(self):
        return """\
ReadJsonOptions
===============
target = {}
lines_to_analyze = {}
encoding = {}
names = {}
detect_data_types = {}""".format(
            self.target,
            self.lines_to_analyze,
            self.encoding,
            self.names,
            self.detect_data_types,
        )
