# Copyright (c) Microsoft Corporation.  All rights reserved.

from ._utils import (
    record_telemetry,
    record_telemetry_reuse_logger,
    record_derived_telemetry,
)
from .._read_parquet_result import ReadParquetResultImpl


class ReadParquetResult(object):
    """"A result object for returning the result of learning code to read parquet file."""

    @record_telemetry_reuse_logger
    def __init__(self, logger, filename, options):
        """Creates a result object for learning code to read a parquet file into a table.
         This should only be called internally by the library.

        :param filename: The parquet file to read the table from.
        :type filename: str
        :param options: Options to specify extra settings to learn code.
        :type options: :class `prose.codeaccelerator.ReadParquetOptions`
        """

        self._impl = ReadParquetResultImpl(logger, filename, options)

    @property
    @record_telemetry(accessor="get")
    def code(self):
        """Return the code that was learned.

        :return: The string representation of synthesized code snippet.
        :rtype: str
        """
        return self._impl.code

    @property
    @record_telemetry(accessor="get")
    def options(self):
        """ReadParquetOptions object containing values that affect the generated code.

        :return:
        :rtype: :class:`prose.codeaccelerator.ReadParquetResult`
        """
        return self._impl.options

    @property
    @record_telemetry(accessor="get")
    def data(self):
        """Return the data obtained when running the synthesized program on the example file.

        :return: The dataframe containing the data. :rtype: Union[`DataFrame
        <https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.html>`_, List[`Row
        <http://spark.apache.org/docs/latest/api/python/pyspark.sql.html#pyspark.sql.Row>`_]]
        """
        return self._impl.data

    @record_telemetry
    def update(self):
        """Attempts to learn a program using the current options."""

        return self._impl.learn()

    @record_telemetry
    def __repr__(self):
        """Return a printable string that summarizes the ReadParquetResult and the learned meta-data."""

        return self._impl.__repr__()
