# Copyright (c) Microsoft Corporation.  All rights reserved.

from prose.codeaccelerator import _internals
from ._read_options import ReadOptions
from .._buffered_reader import ReadTextBufferedReader
from .._config import config


class ReadTextOptions(ReadOptions):
    """Options specific to reading text files."""

    @property
    def _internal(self):
        return self._read_text_options

    def __init__(
        self,
        target=None,
        lines_to_analyze=None,
        program_type=None,
        skip=None,
        delimiter=None,
        schema_filename=None,
        colspecs=None,
        detect_data_types=False,
    ):
        """Initializes options for reading parquet file.

        :param target: Target platform for generated code. Supported values are :code:`"pandas"` and :code:`"pyspark"`.
        :param lines_to_analyze: Number of lines of the input file that should be used during learning.
        :param program_type: Type of the program to read the text file. Supported values are :code:`"csv"` and :code:`"fw"`.
        :param skip: Number of skipped lines in the input file.
        :param delimiter: Single delimiter that the learned program will use to split columns. Valid only for the :code:`"csv"` program type.
        :param schema_filename: Fixed-width schema file to be used for learning. Valid only for the :code:`"fw"` program type.
        :param colspecs: List of pairs of the extents of fixed-width fields of each line.
        :param detect_data_types: Whether or not to synthesize code to detect data types as part of reading the file.
        """
        self._read_text_options = _internals.ReadTextOptions()

        self.target = target if target else ReadTextOptions._default_target()
        if lines_to_analyze is not None:
            self.lines_to_analyze = lines_to_analyze
        if schema_filename:
            self.schema_filename = schema_filename
        if detect_data_types:
            self.detect_data_types = detect_data_types
        if program_type:
            self.program_type = program_type
        if skip is not None:
            self.skip = skip
        if delimiter is not None:
            self.delimiter = delimiter
        if colspecs:
            self.colspecs = colspecs

    @classmethod
    def _default_target(cls):
        """Default target to use for this option."""
        return config.get("default_target", "pandas")

    def _allowed_targets(self):
        """Specify a list of the allowed targets for this builder type."""
        return self._read_text_options.allowed_targets

    @property
    def target(self):
        """Target runtime library for the generated code.

        Supported values are :code:`"pandas"` and :code:`"pyspark"`.

        :type: str
        """
        return self._read_text_options.target

    @target.setter
    def target(self, value):
        value = super()._validate_target(value)
        self._read_text_options.target = value

    @property
    def lines_to_analyze(self):
        """Number of lines of the input file that should be used during learning.

        :type: int
        """
        return self._read_text_options.lines_to_analyze

    @lines_to_analyze.setter
    def lines_to_analyze(self, value):
        if value < 1:
            raise ValueError("The number of lines to analyze must be at least 1.")
        self._read_text_options.lines_to_analyze = value

    @property
    def program_type(self):
        """Type of the program to read the text file.

        Supported values are :code:`"csv"` and :code:`"fw"`.

        :type: str
        """
        return self._read_text_options.program_type

    @program_type.setter
    def program_type(self, value):
        self._read_text_options.override_program_type(value)

    @property
    def encoding(self):
        """The detected encoding of the sample file that will be used for subsequent reads.

        :type: str
        """
        return self._read_text_options.encoding

    @property
    def names(self):
        """List of column names learned from the input file that will be used in the output table.

        :type: List[str]
        """
        return self._read_text_options.names

    @property
    def skip(self):
        """Number of skipped lines in the input file.

        :type: int
        """
        return self._read_text_options.skip

    @skip.setter
    def skip(self, value):
        self._read_text_options.override_skip(value)

    @property
    def delimiter(self):
        """Single delimiter that the learned program will use to split columns if one exists; None otherwise.

        :type: str
        """
        return self._read_text_options.delimiter

    @delimiter.setter
    def delimiter(self, value):
        self._read_text_options.override_delimiter(value)

    @property
    def schema_filename(self):
        """Fixed-width schema file to be used for learning.

        :type: str
        """
        return self._read_text_options.schema_filename

    @schema_filename.setter
    def schema_filename(self, value):
        if value is None:
            self._schema_reader = None
        else:
            # Storing value on Python side as well as otherwise it gets GCd
            self._schema_reader = ReadTextBufferedReader(value, create_stream=True)
        self._read_text_options.set_schema_file(value, self._schema_reader)

    @property
    def colspecs(self):
        """
        List of pairs of the extents of fixed-width fields of each line as half-open intervals (i.e., [from, to[ ).

        :type: List[Tuple[int, int]]
        """
        colspecs = self._read_text_options.colspecs
        if colspecs is None:
            return None

        return [(cs.start, cs.end) for cs in self._read_text_options.colspecs]

    def to_serializable(self):
        return {
            "target": self.target,
            "lines_to_analyze": self.lines_to_analyze,
            "program_type": self.program_type,
            "skip": {
                "value": self.skip,
                "overridden": self._read_text_options.skip_overridden,
            },
            "delimiter": {
                "value": self.delimiter,
                "overridden": self._read_text_options.delimiter_overridden,
            },
            "schema_filename": self.schema_filename,
            "colspecs": self.colspecs,
            "detect_data_types": self.detect_data_types,
        }

    @classmethod
    def from_serializable(cls, serializable):
        skip = serializable.get("skip")
        delimiter = serializable.get("delimiter")
        return cls(
            serializable.get("target"),
            serializable.get("lines_to_analyze"),
            serializable.get("program_type"),
            skip.get("value") if skip is not None and skip.get("overridden") else None,
            delimiter.get("value")
            if delimiter is not None and delimiter.get("overridden")
            else None,
            serializable.get("schema_filename"),
            serializable.get("colspecs"),
            serializable.get("detect_data_types"),
        )

    def __repr__(self):
        # TODO: This looks ugly when part of ReadTextResult output
        return """\
ReadTextOptions
===============
target = {}
lines_to_analyze = {}
program_type = {}
encoding = {}
names = {}
skip = {}
delimiter = {}
schema_filename = {}
colspecs = {}
detect_data_types = {}""".format(
            self.target,
            self.lines_to_analyze,
            self.program_type,
            self.encoding,
            self.names,
            self.skip,
            self.delimiter,
            self.schema_filename,
            self.colspecs,
            self.detect_data_types,
        )
