# Copyright (c) Microsoft Corporation.  All rights reserved.

from ._shared_utils import create_mimebundle
from ._utils import record_telemetry, record_telemetry_reuse_logger
from ._read_text_options import ReadTextOptions
from .._read_text_result import ReadTextResultImpl


class ReadTextResult(object):
    """"A result object for returning the result of learning code to read text file."""

    @record_telemetry_reuse_logger
    def __init__(self, logger, filename, options):
        """Creates a result object for learning code to read a text file into a table.
         This should only be called internally by the library.

        :param filename: The text file to read the table from.
        :type filename: str
        :param options: Options to specify extra settings to learn code.
        :type options: :class `prose.codeaccelerator.ReadTextOptions`
        """

        self._filename = filename
        self._impl = ReadTextResultImpl(logger, filename, options)

    @property
    @record_telemetry(accessor="get")
    def code(self):
        """Return the code that was learned.

        :return: The string representation of synthesized code snippet.
        :rtype: str
        """
        return self._impl.code

    @property
    @record_telemetry(accessor="get")
    def options(self):
        """ReadTextOptions object containing values that affect the generated code.

        :return:
        :rtype: :class:`prose.codeaccelerator.ReadTextResult`
        """
        return self._impl.options

    @property
    @record_telemetry(accessor="get")
    def data(self):
        """Return the data obtained when running the synthesized program on the example file.

        :return: The dataframe containing the data. :rtype: Union[`DataFrame
        <https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.html>`_, List[`Row
        <http://spark.apache.org/docs/latest/api/python/pyspark.sql.html#pyspark.sql.Row>`_]]
        """
        return self._impl.data

    @record_telemetry
    def update(self):
        """Attempts to learn a program using the current options."""
        self._impl.learn()

    def to_serializable(self):
        return {
            "result_type": "ReadText",
            "file_name": self._filename,
            "code": self.code,
            "data_frame_json": self.data.to_json(orient="table", index=False),
            "options": self.options.to_serializable(),
        }

    @classmethod
    def from_serializable(cls, logger, serializable):
        return cls(
            logger,
            serializable.get("file_name"),
            ReadTextOptions.from_serializable(serializable.get("options")),
        )

    @record_telemetry
    def _repr_mimebundle_(self, include, exclude):
        """Return an object passed to the mime type renderer in the client side extensions."""
        return create_mimebundle(self)

    @record_telemetry
    def __repr__(self):
        """Return a printable string that summarizes the ReadTextResult and the learned meta-data."""
        return self._impl.__repr__()
