# Copyright (c) Microsoft Corporation.  All rights reserved.

# Our lab package uses a separate C# process than the core. That means that types
# exported by the core can't be used in lab. Specifically, each Python package must
# use its own PythonLoggerFactory. The create_* methods here are parameterized by
# the factory type and return a decorator that will work with that type.

import functools
from .._utils import _telemetry_settings

_sessionTypeKey = "SessionType"
_pythonApiKey = "PythonApi"

_proseoutput_v1_mimetype = "application/vnd.proseoutput.v1+json"


def _session_properties(session_type, method, accessor=None):
    method = method + "." + accessor if accessor is not None else method
    return [(_sessionTypeKey, session_type), (_pythonApiKey, method)]


def _clear_stackframe_variables(e):
    # Clear the local variables from all stack frames present in the current
    # traceback object
    # See: https://cosmicpercolator.com/2016/01/13/exception-leaks-in-python-2-and-3/
    import traceback

    traceback.clear_frames(e.__traceback__)


def create_new_telemetry(python_logger_factory_class):
    def _create_new_telemetry(func):
        """Decorator that creates a new telemetry logger and records telemetry for a
        function call. The wrapped method should return a lambda over a logger.
        This decorator will apply that lambda to the new logger and leave the result
        of the lambda as the actual function return value.

        :param func: Function to be wrapped.
        """

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            class_name = (
                args[0].__name__  # for @classmethod
                if isinstance(args[0], type)
                else args[0].__class__.__name__  # for instance methods
            )
            logger = python_logger_factory_class().new_base(
                _telemetry_settings(), class_name
            )
            properties = _session_properties(class_name, func.__name__)
            logger.track_event(properties)
            try:
                result = func(*args, **kwargs)(logger)
            except Exception as e:
                logger.track_exception(properties)

                # Setting logger to None, removes the reference to background process
                # which can then get collected.
                logger = None
                _clear_stackframe_variables(e)
                raise

            return result

        return wrapper

    return _create_new_telemetry


def create_derived_telemetry(python_logger_factory_class):
    """Function that is parameterized by a factory class to begin the chain of calls
    for decorating a method with telemetry actions.

    The following example is the intended usage:
        record_derived_telemetry_for_new_session = create_derived_telemetry(PythonLoggerFactory)
        ...
        @record_derived_telemetry("ReadFwfLearnResult")
        def learn(self):
            ...

    Note that the variable record_derived_telemetry in the example does not need to
    have the same name as the function below, but making them the same helps clarify
    which of these functions are assigned to that variable. The result of applying
    record_derived_telemetry to a name ("ReadFwfLearnResult" in this example) is a
    decorator, which then modifies the function ("learn" in this example) to perform
    telemetry.
    """

    def _create_derived_telemetry(session_type):
        """Function that accepts the new session type name for logging and returns
        a decorator for recording telemetry with that new name.
        """

        def decorator(func):
            """Decorator that creates a derived telemetry logger and records telemetry
            for a function call. The existing logger will be extracted from the first
            argument to the wrapped method. The wrapped method should return a lambda
            over a logger that will be evaluated by this decorator. This decorator will
            apply that lambda to the new logger and leave the result of the lambda as
            the actual function return value.

            :param func: Function to be wrapped.
            """

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                class_name = args[0].__class__.__name__
                old_logger = args[0]._impl._internal.python_logger
                new_logger = python_logger_factory_class().new_derived(
                    old_logger, session_type
                )
                new_id = new_logger.correlation_string
                properties = _session_properties(class_name, func.__name__)
                properties.append(("ChildId", new_id))
                old_logger.track_event(properties)
                try:
                    result = func(*args, **kwargs)(new_logger)
                except Exception as e:
                    new_logger.track_exception(properties)

                    # Setting logger to None, removes the reference to background process
                    # which can then get collected.
                    new_logger = None
                    _clear_stackframe_variables(e)
                    raise

                return result

            return wrapper

        return decorator

    return _create_derived_telemetry


def record_telemetry_for_builder_api(func=None, accessor=None):
    """Decorator that records telemetry for a function.

    This must be applied before @property because it
    operates on the underlying function, not the property.

    :param func: Function to be wrapped.
    :param accessor: An optional component of the function to be specified
    (e.g., get or set for properties).
    """
    if callable(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            properties = _session_properties(
                args[0].__class__.__name__, func.__name__, accessor=accessor
            )
            args[0]._impl._internal.python_logger.track_event(properties)
            try:
                return func(*args, **kwargs)
            except:
                args[0]._impl._internal.python_logger.track_exception(properties)
                raise

        return wrapper
    else:
        return functools.partial(record_telemetry_for_builder_api, accessor=accessor)


def create_mimebundle(obj):
    """Create a mimebundle representing the given learn result object

    :param obj: The object to represent.
    :type obj: object (learn result)
    :return: The mimebundle dict.
    :rtype: dict
    """
    try:
        # try to_serializable first
        return {_proseoutput_v1_mimetype: obj.to_serializable()}
    except Exception:
        # return the string representation
        return {"text/plain": str(obj)}
