# Copyright (c) Microsoft Corporation.  All rights reserved.

# Our lab package uses a separate C# process than the core. That means that types
# exported by the core can't be used in lab. Specifically, each Python package must
# use its own PythonLoggerFactory. The create_* methods here are parameterized by
# the factory type and return a decorator that will work with that type.

import functools

from .._utils import _telemetry_settings
from ._shared_utils import _clear_stackframe_variables

_pythonClassKey = "PythonClass"
_pythonApiKey = "PythonApi"


def _class_properties(method_name, class_name=None, accessor=None):
    method_name = method_name + "." + accessor if accessor is not None else method_name
    props = [(_pythonApiKey, method_name)]
    if class_name is not None:
        props += [(_pythonClassKey, class_name)]
    return props


def create_new_telemetry_for_top_level(python_logger_factory_class):
    def _create_new_telemetry_for_top_level(func):
        """Decorator that creates a new telemetry logger for top level API invocation and records
        telemetry for a function call. The wrapped method should return a lambda over a logger.
        This decorator will apply that lambda to the new logger and leave the result of the lambda
        as the actual function return value.

        This decorator should be used on APIs (vs. class methods or constructors) that will be directly
        invoked by the user.

        :param func: Function to be wrapped.
        """

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            logger = python_logger_factory_class().new_base(
                _telemetry_settings(), func.__name__
            )
            logger.track_event(_class_properties(func.__name__))
            try:
                result = func(*args, **kwargs)(logger)
            except Exception as e:
                logger.track_exception(_class_properties(func.__name__))

                # Setting logger to None, removes the reference to background process
                # which can then get collected.
                logger = None
                _clear_stackframe_variables(e)
                raise

            return result

        return wrapper

    return _create_new_telemetry_for_top_level


def record_telemetry_reuse_logger(func):
    """Decorator that passes the existing logger to the function that is being decorated.
    The wrapped method should take logger as 2nd parameter (1st being self) that it can set
    internally and further use for logging

    :param func: Function to be wrapped.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        python_logger = args[1]
        properties = _class_properties(
            func.__name__, class_name=args[0].__class__.__name__
        )
        python_logger.track_event(properties)
        try:
            return func(*args, **kwargs)
        except Exception as e:
            python_logger.track_exception(properties)

            # Setting python_logger to None, removes the reference to background process
            # which can then get collected.
            python_logger = None
            _clear_stackframe_variables(e)
            raise

    return wrapper


def record_telemetry(func=None, accessor=None):
    """Decorator that records telemetry for a function.

    This must be applied before @property because it
    operates on the underlying function, not the property.

    :param func: Function to be wrapped.
    :param accessor: An optional component of the function to be specified
    (e.g., get or set for properties).
    """
    if callable(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            properties = _class_properties(
                func.__name__, class_name=args[0].__class__.__name__, accessor=accessor
            )
            args[0]._impl._internal.python_logger.track_event(properties)
            try:
                return func(*args, **kwargs)
            except:
                args[0]._impl._internal.python_logger.track_exception(properties)
                raise

        return wrapper
    else:
        return functools.partial(record_telemetry, accessor=accessor)
