# Copyright (c) Microsoft Corporation.  All rights reserved.
from ._telemetry_utils import (
    create_new_telemetry_for_top_level,
    record_telemetry_reuse_logger,
    record_telemetry,
)
from ._shared_utils import (
    create_derived_telemetry,
    create_new_telemetry,
    record_telemetry_for_builder_api,
)
from .._internals import PythonLoggerFactory

record_new_telemetry = create_new_telemetry(PythonLoggerFactory)
record_derived_telemetry = create_derived_telemetry(PythonLoggerFactory)
record_new_telemetry_for_top_level = create_new_telemetry_for_top_level(
    PythonLoggerFactory
)

import pandas as pd
import json
import datetime


def _serialize_dataframe(df):
    """Serialize the given pandas Dataframe object."""
    # use df.to_json(orient='table', index=False) to do the serialization.
    json_str = df.to_json(orient="table", index=False)
    # to_json maps date and time objects to strings and sets their type to string.
    # So, we have to reset the type of columns that have dates, times as 'date' and 'time'
    df_json_obj = json.loads(json_str)
    for t in df_json_obj["schema"]["fields"]:
        # df.to_json correctly sets the type for int, float, and datetime columns
        if t["type"] in ["integer", "number", "datetime"]:
            continue
        # df.to_json incorrectly sets the type for date and time columns to 'string'
        # we are assuming that the first row has index 0
        if type(df[t["name"]][0]) == datetime.date:
            t["type"] = "date"
            continue
        if type(df[t["name"]][0]) == datetime.time:
            t["type"] = "time"
            continue
    return json.dumps(df_json_obj)


def _deserialize_dataframe(serialized_dataframe):
    """Assuming serialized_dataframe is created using df.to_json(orient='table', index=False),
       this function deserializes the object back to a pandas df"""
    df_json_obj = json.loads(serialized_dataframe)
    # Get the correct order of columns; order is preserved in json.schema.fields.
    columns = [f["name"] for f in df_json_obj["schema"]["fields"]]
    # df_json_obj.data stores rows as a dict from str(column name) to value.
    df = pd.DataFrame.from_records(df_json_obj["data"], columns=map(str, columns))
    # rename the columns to their original names
    df.columns = columns

    # type cast the values to whatever types they were
    for t in df_json_obj["schema"]["fields"]:
        if t["type"] == "integer":
            df[t["name"]] = pd.Series(
                df[t["name"]].apply(lambda x: int(x)), dtype=object
            )
            continue
        if t["type"] == "number":
            df[t["name"]] = pd.Series(
                df[t["name"]].apply(lambda x: float(x)), dtype=object
            )
            continue
        # either datetime or date or time or string
        if t["type"] == "datetime":
            df[t["name"]] = pd.to_datetime(df[t["name"]])
            continue
        # df.to_json (when orient=table) maps date and time ojects to strings in ISO8601 format
        # so we assume this format when parsing the date and time strings
        if t["type"] == "date":
            df[t["name"]] = df[t["name"]].apply(
                lambda x: datetime.date(int(x[0:4]), int(x[5:7]), int(x[8:10]))
            )
            continue
        if t["type"] == "time":
            df[t["name"]] = df[t["name"]].apply(
                lambda x: datetime.time(
                    int(x[0:2]),
                    int(x[3:5]),
                    int(x[6:8]),
                    int(x[10:]) if len(x) > 19 else 0,
                )
            )
            continue
        # we reach here if column was 'string', but there is nothing to do in that case
    return df
