# Copyright (c) Microsoft Corporation.  All rights reserved.

from abc import ABC, abstractmethod
from . import _internals
from ._buffered_reader import ReadTextBufferedReader
from ._config import config


class Base(ABC):
    @property
    @abstractmethod
    def _internal(self):
        """Get the C# internal object that is backing this object."""


class BuilderImpl(Base):
    @property
    def _internal(self):
        return self._builder

    @property
    def _telemetry_testing_data(self):
        """Get the list of collected telemetry events. None, if not in telemetry testing mode."""
        return (
            self._builder.get_telemetry_testing_data()
            if self._builder is not None
            else None
        )

    @classmethod
    @abstractmethod
    def _default_target(cls):
        """Specify the default target for this builder type."""

    @classmethod
    @abstractmethod
    def _allowed_targets(cls):
        """Specify a list of the allowed targets for this builder type."""

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, value):
        lower_value = self._validate_target(value)
        self._target = lower_value

    def _validate_target(self, value):
        lower_value = value.lower()
        if lower_value not in self.__class__._allowed_targets():
            raise ValueError(
                "Invalid value '{}' for 'target'. Expected a value from {}.".format(
                    value, self.__class__._allowed_targets()
                )
            )
        return lower_value


class LearnResultImpl(Base):
    @property
    def _internal(self):
        return self._result


class ReadTableBuilderImpl(BuilderImpl):
    def __init__(self, filename):
        self._filename = filename
        self._file_reader = ReadTextBufferedReader(filename, create_stream=True)

    @classmethod
    def _default_target(cls):
        return config.get("default_target", "pandas")

    @classmethod
    def _allowed_targets(cls):
        return ["pandas", "pyspark"]

    @BuilderImpl.target.setter
    def target(self, value):
        # Incantation from https://stackoverflow.com/a/37663266
        super(ReadTableBuilderImpl, self.__class__).target.fset(self, value)
        self._builder.target = self.target

    @property
    def filename(self):
        return self._filename

    @property
    def lines_to_analyze(self):
        return self._builder.lines_to_analyze

    @lines_to_analyze.setter
    def lines_to_analyze(self, value):
        if value is None:
            self._builder.lines_to_analyze = value
        elif not isinstance(value, int):
            raise ValueError("The number of lines to analyze must be integer.")
        elif value < 1:
            raise ValueError("The number of lines to analyze must be at least 1.")
        else:
            self._builder.lines_to_analyze = value


class ReadTableLearnResultImpl(LearnResultImpl):
    @property
    def column_count(self):
        return self._result.column_count

    @property
    def names(self):
        return self._result.names
