# Copyright (c) Microsoft Corporation.  All rights reserved.

import pandas as pd
import numpy as np
from datetime import datetime, date, time
from ._internals import CellValueFactory


class CellValueConverter(object):
    # Cached type for pandas's Timestamp
    _timestamp_type = type(pd.Timestamp(2019, 1, 1))

    # Factory to create CellValue objects.
    _factory = CellValueFactory()

    @classmethod
    def _normalize_type(cls, type_to_check):
        """
        Normalize the data types detected by pandas's data frame and map them back to
        Python.
        Currently the only data type it normalize is pandas's Timestamp. If a data frame column
        contains all datetime values, data frame's cell data's type is Timestamp. However, if one
        of the row contains string along with other datetime values, the type of cell data returned
        is datetime. This method would check if the type was Timestamp and if yes, returns datetime.
        For all other cases, it will return the type as is.

        :param type_to_check: Type of data present in panda's data frame cell.
        :return: datetime, if type_to_check is Timestamp, otherwise type_to_check.
        """
        if type_to_check in [cls._timestamp_type, date, time, np.datetime64]:
            return datetime
        if issubclass(type_to_check, np.integer):
            return int
        if issubclass(type_to_check, np.floating):
            return float
        return type_to_check

    @classmethod
    def _is_none(cls, data_to_check):
        """
        Checks if data is one of the NA values as defined in pandas's `isna` method.
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.isna.html

        :param data_to_check: Data to check for.
        :return: True if data is none, else False.
        """
        return pd.isna(data_to_check)

    @classmethod
    def _create_cell_value(cls, data, expected_type):
        """
        Create an instance of `CellValue` passing the `data` as the underlying value.
        It validates if the type of data matches the `expected_type`. If not, raise ValueError.

        :param data: Data for which CellValue need to be created.
        :param expected_type: Expected type of `data`.
        :return: CellValue object
        """
        data_type = cls._normalize_type(type(data))

        if not cls._is_none(data) and expected_type != data_type:
            raise ValueError(
                'Expected data of type "%s", but found that of "%s".'
                % (expected_type.__name__, data_type.__name__)
            )

        if expected_type == int:
            # surprisingly, we don't need to do "data = data.item()" when data is numpy.int/float
            return cls._factory.create_int_value(data)
        elif expected_type == float:
            # For None, pandas convert it to nan, so make sure we store None and not nan.
            data = data if not cls._is_none(data) else None
            return cls._factory.create_double_value(data)
        elif expected_type == str:
            # pandas read_csv maps string "NULL" to python float "NaN"
            data = None if isinstance(data, float) else data
            return cls._factory.create_string_value(data)
        elif expected_type == datetime:
            # first, if data is datetime.time then convert it to python datetime
            # pd.to_datetime can take datetime or date as input, but not time.
            data = (
                datetime.combine(date(1970, 1, 1), data) if type(data) == time else data
            )
            # For None, pandas convert it to NaT, so make sure we store None and not NaT.
            data = pd.to_datetime(data) if not cls._is_none(data) else None
            return cls._factory.create_date_time_value(data)
        else:
            raise ValueError(
                'Unsupported data "%s" of type "%s"' % (data, data_type.__name__)
            )

    @classmethod
    def _get_cell_value(cls, cell_value, expected_type):
        """
        Reads value from `cell_value`. The method refers to `expected_type` to determine
        how to fetch the right type of value from CellValue.

        :param cell_value: CellValue object for which underlying value need to be fetched.
        :param expected_type: The underlying python value of CellValue.
        :return:
        """
        if expected_type == int:
            return cell_value.int_value
        elif expected_type == float:
            return cell_value.double_value
        elif expected_type == str:
            return cell_value.string_value
        elif expected_type == datetime:
            return cell_value.date_time_value
        else:
            raise ValueError(
                'Unsupported expected data type "%s", to convert data of type "%s".'
                % (expected_type.__name__, cell_value.type)
            )

    @classmethod
    def tuple_to_cell_values(cls, input_tuple, expected_types):
        """
        Converts a tuple containing python data into list of CellValue. The method ensures the type
        of all the entries inside `input_tuple` matches to that present in `expected_types`. If not,
        will raise ValueError.

        :param input_tuple: Tuple to convert.
        :param expected_types: List of expected type for every entry of `input_tuple`.
        :return: List of CellValue corresponding to the `input_tuple`.
        """
        if len(input_tuple) != len(expected_types):
            raise ValueError(
                "Mismatch in entries present in input_tuple vs. expected_types."
            )

        result = []
        cell_index = 0
        for data in input_tuple:
            result.append(cls._create_cell_value(data, expected_types[cell_index]))
            cell_index = cell_index + 1
        return result

    @classmethod
    def cell_values_to_tuple(cls, list_of_cell_values, expected_types):
        """
        Converts a list of CellValue into tuple. The method refers to `expected_type` to determine
        how to fetch the right type of value from CellValue.

        :param list_of_cell_values: List of CellValue to convert.
        :param expected_types: Expected type of cell values.
        :return: Tuple containing the underlying value present in `list_of_cell_values`.
        """
        if len(list_of_cell_values) != len(expected_types):
            raise ValueError(
                "Mismatch in entries present in list_of_cell_values vs. expected_types."
            )

        result = []
        cell_index = 0
        for cell_value in list_of_cell_values:
            result.append(cls._get_cell_value(cell_value, expected_types[cell_index]))
            cell_index = cell_index + 1
        return tuple(result)

    @classmethod
    def data_frame_to_cell_values(cls, input_dataframe):
        """
        Converts the input data to list of list of CellValue.

        :param input_dataframe: Input data frame
        :return: List of list of CellValue.
        """

        def _get_column_types(data_frame):
            """
            Returns a list of types of data present in every column by inspecting the first non None
            value of the column and storing its type.

            :param data_frame: Data frame on which column types need to be known.
            :return: List of types corresponding to no. of columns in the data frame.
            """
            col_types = []
            for curr_col_index in data_frame.columns:
                curr_column = data_frame[curr_col_index]
                non_none_value = next(
                    (value for value in curr_column if not cls._is_none(value)), None
                )
                col_types.append(cls._normalize_type(type(non_none_value)))
            return col_types

        result = []

        # Get the type of data present in every column
        column_types = _get_column_types(input_dataframe)

        # Iterate over rows/columns and convert every cell to CellValue.
        for row in input_dataframe.iterrows():

            # 0 - column id, 1 - data
            curr_col = row[1]
            column_index = 0
            col_result = []
            for input_data in curr_col:
                col_result.append(
                    cls._create_cell_value(input_data, column_types[column_index])
                )
                column_index = column_index + 1

            result.append(col_result)

        return result, column_types
