# Copyright (c) Microsoft Corporation.  All rights reserved.
import ast

from ._utils import executes_generated_code


class CodeSnippet(object):
    def __init__(self, code_string):
        assert code_string is not None
        self.code_string = code_string
        self._compiled_code_block = None
        self._compiled_last_expr = None

    def _compile_code(self):
        """
        Compiles the code that represents this instance of :CodeSnippet:.
        It splits the code in multiple blocks and then checks if the last block is an Expression.
        If yes, it will compile all the blocks but last using mode='exec' and compile the last expression block
        using mode='eval' whose result will be returned when the code snippet will get evaluated.

        If last block was not an expression, all the blocks are just compiled using mode='exec'.

        It inserts the key/value pairs of kwargs in the context of code during execution.
        :return:
        """
        code_block_ast = ast.parse(self.code_string, mode="exec")
        last_block_ast = code_block_ast.body[-1]

        if isinstance(last_block_ast, ast.Expr):
            # If last block is an expression, pop it from main code-block
            code_block_ast.body.pop()
            last_expr_ast = ast.Expression(last_block_ast.value)
            self._compiled_last_expr = compile(last_expr_ast, "<string>", mode="eval")

        self._compiled_code_block = compile(code_block_ast, "<string>", mode="exec")

    @executes_generated_code
    def evaluate(self, **kwargs):
        """
        Executes the compiled code and returns the result.

        It inserts the key/value pairs of kwargs in the context of code during execution.
        :return:
        """
        if self._compiled_code_block is None:
            self._compile_code()

        _globals = kwargs

        lambda_body = (exec(self._compiled_code_block, _globals),)
        if self._compiled_last_expr is not None:
            lambda_body += (eval(self._compiled_last_expr, _globals),)

        # Execute the code inside a lambda and return the result of last statement
        return (lambda: lambda_body[-1])()

    def __repr__(self):
        return self.code_string
