# Copyright (c) Microsoft Corporation.  All rights reserved.

import collections.abc
import json
import os

from pathlib import Path

__all__ = ["config"]


def env(key, default=None):
    """
    Given an environment variable, return the value as lowercase.

    :param str key: The configuration key.
    :param default: The value to return if the key is not found (defaults to None).
    :return The lowercased value of the environment variable if found, otherwise default.
    """
    key = key.upper()
    value = os.environ.get(key, default)
    if value is not None:
        value = value.lower()

    return value


class Config(collections.abc.MutableMapping):
    """Global configuration for PROSE."""

    def __init__(self):
        # Initialize dictionary with default key/value pairs.
        # These have the lowest priority.
        self._defaults = {
            # No signal means that the user is consenting to sending telemetry by default.
            "telemetry_optout": False,
            "telemetry_testing_mode": False,
            "host_name": None,
        }
        self._dict = dict(self._defaults)

        # These keys cannot be set externally.
        self._readonly = frozenset(["host_name"])

        # If a configuration file exists, these take priority
        # over defaults.
        prose_config_path = Path.home() / ".config/prose/config.json"
        if prose_config_path.exists():
            with prose_config_path.open() as f:
                try:
                    for k, v in json.load(f).items():
                        self[k] = v
                except ValueError as e:
                    raise ValueError(
                        "Configuration file {} contains invalid JSON.".format(
                            prose_config_path
                        )
                    ) from e

        # Environment variables have the highest priority.
        mshost_telemetry_env = env("MSHOST_TELEMETRY_ENABLED")
        prose_telemetry_optout_env = env("PROSE_TELEMETRY_OPTOUT")

        # If MSHOST_TELEMETRY_ENABLED exists, but the key isn't explicitly set to true, optout.
        if mshost_telemetry_env and mshost_telemetry_env != "true":
            self._dict["telemetry_optout"] = True

        # If PROSE_TELEMETRY_OPTOUT exists, optout.
        if prose_telemetry_optout_env:
            self._dict["telemetry_optout"] = True

        # Update telemetry_optout default based on new signals.
        self._defaults["telemetry_optout"] = self._dict["telemetry_optout"]

        # MSHOST_ENVIRONMENT is of the form <program> or <program>-<version>.
        host_env = env("MSHOST_ENVIRONMENT")
        if host_env:
            # name, sep (unused), version (unused)
            name, _, _ = host_env.partition("-")
            if name in ("automl", "adsclient", "adscluster"):
                self._dict["host_name"] = name

    @property
    def readonly(self):
        return self._readonly

    def _save(self):
        # To reload for testing, restore when done.
        class old:
            def __init__(self, config):
                self._config = config
                self._old_defaults = config._defaults
                self._old_dict = config._dict
                config._defaults = dict(config._defaults)
                config._dict = dict(config._dict)

            def __enter__(self):
                return self

            def _reload(self):
                self._config.__init__()

            def __exit__(self, type, value, traceback):
                self._config._defaults = self._old_defaults
                self._config._dict = self._old_dict

        return old(self)

    def __getitem__(self, key):
        return self._dict[key]

    def __setitem__(self, key, value):
        if key in self.readonly:
            raise ValueError("{} is read-only and cannot be set.".format(key))

        self._dict[key] = value

    def __delitem__(self, key):
        if key in self.readonly:
            raise ValueError("{} is read-only and cannot be deleted.".format(key))

        del self._dict[key]

        # If there is a default configuration for this key after
        # deletion, add the default back.
        if key in self._defaults:
            self._dict[key] = self._defaults[key]

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)


config = Config()
