# Copyright (c) Microsoft Corporation.  All rights reserved.

import os
import pandas
import functools
from prose.codeaccelerator import exceptions, _functions, _utils
from ._cell_value_converter import CellValueConverter
from ._exampledict import _ExampleDict

from ._api._derive_value_options import DeriveValueOptions
from . import _internals, _code
from ._result import ResultImpl


def _run_code_in_wrapper(code_func, *arg):
    """Running generated code on additional inputs can generate exceptions. Use this instead."""
    try:
        return code_func(*arg)
    except:
        return None


# Maximum number of rows to use for learning, etc.
MAX_ROWS = 100


class DeriveValueResultImpl(ResultImpl):
    @property
    def _internal(self):
        return self

    @property
    def python_logger(self):
        return self._python_logger

    def __init__(self, logger, input_table, options):
        self._python_logger = logger
        self._input_table_type = type(input_table)
        self._df = pandas.DataFrame(input_table)
        self._sampled_df = self._df.head(MAX_ROWS)
        # TODO Eventually _sampled_df should be diversity-sampled from _df
        self._options = options or DeriveValueOptions()
        self._options.input_column_names = self._options.input_column_names or list(
            self._sampled_df.columns
        )
        self._options.examples = _ExampleDict(logger, self._options.examples)

        self._column_types = None
        self._generated_function = None
        self._code_string = None
        self._code_snippet = None
        self._data = self._df.copy(deep=False)
        self.learn()

    @property
    def code(self):
        return self._code_string

    @property
    def options(self):
        return self._options

    @property
    def data(self):
        # if data has already been computed once before, return it
        if self.options.new_column_name in self._data.columns:
            return self._data

        df_on_input_columns = (
            self._df
            if self.options.input_column_names is None
            else self._df[self.options.input_column_names]
        )
        self._data[self.options.new_column_name] = [
            _run_code_in_wrapper(self._generated_function, *row)
            for i, row in df_on_input_columns.iterrows()
        ]
        return self._data

    def learn(self):
        if not self.options.examples and not self.options._index_based_examples:
            return

        # remove the column that was added by the last learn call
        if self.options.new_column_name in self._data.columns:
            del self._data[self.options.new_column_name]

        # update the inputs_df, which will be the same as _sampled_df but restricted to input_column_names
        input_column_names = self.options.input_column_names
        inputs_df = self._sampled_df
        if input_column_names:
            if not all(i in self._sampled_df.columns for i in input_column_names):
                raise ValueError(
                    "Invalid values '{}' for 'input_column_names'. All values expected to be in {}.".format(
                        input_column_names, self._sampled_df.columns
                    )
                )
            if set(inputs_df.columns) != set(input_column_names):
                inputs_df = self._sampled_df[input_column_names]

        _inputs_rows, self._column_types = CellValueConverter.data_frame_to_cell_values(
            inputs_df
        )

        # next, collect all the examples
        all_examples = self.options.examples
        for example in self.options._index_based_examples:
            all_examples[self._df.iloc[example["row"], :]] = example["output"]

        examples = []
        for k, v in all_examples.items():
            if self._column_types is None or len(self._column_types) == 0:
                self._column_types = [type(example_value) for example_value in k]
            column_values = CellValueConverter.tuple_to_cell_values(
                k, self._column_types
            )
            examples.append((column_values, v))

        # finally, create the builder and learn.
        builder = _internals.DeriveValueBuilder(self._python_logger, _inputs_rows)
        result_obj = builder.learn(
            self._python_logger, examples, self.options._bound_names
        )
        if result_obj.error is not None:
            raise exceptions.LearnError(result_obj.error)
        # use the result_obj to populate the code and data
        code_string = result_obj.get_code_string()
        self._generated_function = _functions.VarArgFunction(
            "derive_value_program", code_string
        )
        self._code_string = code_string + self._get_code_iter_rows()
        self._code_snippet = _code.CodeSnippet(self._code_string)
        self.options.suggested_examples = self._suggested_examples(
            result_obj.get_suggestions()
        )

    def _get_code_iter_rows(self):
        code = ""
        variable_name = self.options._input_table_variable_name
        if self._input_table_type != pandas.DataFrame:
            code += os.linesep + "import pandas"
            code += os.linesep + "input_table_as_df = pandas.DataFrame({})".format(
                variable_name
            )
            variable_name = "input_table_as_df"
        focused_variable_name = variable_name
        if set(self._sampled_df.columns) != set(self.options.input_column_names):
            focused_variable_name = "{}[[{}]]".format(
                variable_name, ",".join(str(x) for x in self.options.input_column_names)
            )
        code += (
            os.linesep
            + "{0}['{1}'] = [derive_value_program(*row) for i, row in {2}.iterrows()]".format(
                variable_name, self.options.new_column_name, focused_variable_name
            )
        )
        code += os.linesep + "{0}".format(variable_name)
        return code

    def _suggested_all_tuples(self, suggestions):
        """Helper function for suggestions that returns a list with elements of the form
        :code:`((inputs), [outputs])`.  :code:`inputs` may be a singleton but is always
        a :code:`tuple`.  :code:`outputs` is currently just the output from the
        synthesized program.
        """

        result = []
        for suggestion in suggestions:
            row_tuple = CellValueConverter.cell_values_to_tuple(
                suggestion, self._column_types
            )
            suggested_output = _run_code_in_wrapper(
                self._generated_function, *row_tuple
            )
            result.append(
                (row_tuple, [suggested_output] if suggested_output is not None else [])
            )
        return result

    def _suggested_examples(self, suggestions):
        suggested_all_tuples = self._suggested_all_tuples(suggestions)
        suggestions = []
        for inputs, outputs in suggested_all_tuples:
            if len(inputs) == 1:
                suggestions.append((inputs[0], outputs))
            else:
                suggestions.append((inputs, outputs))
        return suggestions

    def __repr__(self):
        suggestions = self.options.suggested_examples or []
        suggestion_text_list = [
            "{} -> {}".format(repr(s[0]), s[1]) for s in suggestions
        ]
        suggestion_text = "\n".join(suggestion_text_list)
        preview_data = (
            [] if self.data is None else self.data[0 : _utils.PREVIEW_DATA_LIMIT]
        )

        return """DeriveValueResult
======================
-------
options
-------
{}

----
code
----
{}

------------
preview_data
------------
{}

------------------
suggested_examples
------------------
{}""".format(
            self.options, self.code, preview_data, suggestion_text
        )
