# Copyright (c) Microsoft Corporation.  All rights reserved.
import math
import os
from collections import OrderedDict

import numpy
import pandas

from .exceptions import ProseException
from ._api._detect_types_options import DetectTypesOptions
from ._result import ResultImpl
from . import _internals, _code, _utils, _iterable_streamer


class _ColumnInfo(object):
    def __init__(self, data, name, true_length):
        if not isinstance(name, str) and not isinstance(name, int):
            raise _utils.ProseException(
                "Column names for DataFrames must be string or integer valued."
            )

        self.data = data
        self.internal_name = name
        self.str_name = str(name)
        self.name_is_int = isinstance(self.internal_name, int)
        self.data_types = list(
            set(type(x) for x in filter(lambda _x: _x is not None, data))
        )
        self.num_values = len(data)
        self.true_length = true_length if true_length is not None else len(self.data)

        if self._should_generate_identity_transform():
            # Don't bother passing data to the native ColumnInfo
            # object if we're not using the column for learning.
            self.data_for_learning = []
        elif self._need_compensation_for_pandas_nan_bug():
            self.data_for_learning = [str(x) for x in self.data]
        else:
            self.data_for_learning = self.data

        # noinspection PyProtectedMember
        self.streamer = _iterable_streamer._IterableStreamer(self.data_for_learning)
        self.native_object = _internals.ColumnInfo(
            self.str_name,
            self.name_is_int,
            self._need_compensation_for_pandas_nan_bug(),
            not self._should_generate_identity_transform(),
            self.streamer,
            self.true_length,
        )

    def _need_compensation_for_pandas_nan_bug(self):
        """See: https://github.com/pandas-dev/pandas/issues/14978"""
        if (
            self.num_values > 0
            and len(self.data_types) == 2
            and str in self.data_types
            and any("float" in str(x) for x in self.data_types)
        ):
            # check the data
            non_string_values = list(
                set(filter(lambda x: not isinstance(x, str), self.data))
            )
            return (
                len(non_string_values) == 1
                and str(non_string_values[0]).lower() == "nan"
            )
        return False

    def _should_generate_identity_transform(self):
        if (
            self.num_values > 0
            and len(self.data_types) == 1
            and self.data_types[0] == str
        ):
            return False
        return not self._need_compensation_for_pandas_nan_bug()


class DetectTypesResultImpl(ResultImpl):
    @property
    def _internal(self):
        return self._result

    def __init__(self, logger, data, options, input_variable_name="input_data"):
        self._options = DetectTypesOptions() if options is None else options
        self._code_snippet = None
        self._code_string = None
        self._input_variable_name = input_variable_name

        type_of_data = type(data)
        self._input_data = data

        true_length = self._options._internal.true_length

        if type_of_data is pandas.DataFrame:
            mode = "dataframe"
            column_infos = [_ColumnInfo(data[x], x, true_length) for x in data.columns]
        elif isinstance(data, dict):
            mode = "dictionary"
            column_infos = [
                _ColumnInfo(data[column_name], column_name, true_length)
                for column_name in sorted(data.keys())
            ]
        else:
            try:
                data_as_list = list(data)
                mode = "list"
                column_infos = [_ColumnInfo(data_as_list, "Column_1", true_length)]
            except Exception:
                raise TypeError(
                    "The input data must be a pandas DataFrame, "
                    "a dictionary with string-valued lists as values "
                    "or a non-nested list of string values."
                )
        self._options._set_mode(mode)

        # Need to pin the streamers for all the columns so when request for streaming comes
        # from C#, streaming is still possible and the iterator doesn't get GC.
        self._data_streamer_for_columns = [
            col_info.streamer for col_info in column_infos
        ]

        self._result = _internals.DetectTypesResult(
            logger, [x.native_object for x in column_infos]
        )
        self.learn()

    @property
    def code(self):
        return self._code_string

    @property
    def options(self):
        return self._options

    @property
    def data(self):
        if self.options.target == "pyspark":
            raise ProseException("Cannot get data when target==pyspark.")

        if self._code_snippet is None:
            return None

        code_snippet_args = {self._input_variable_name: self._input_data}
        return self._code_snippet.evaluate(**code_snippet_args)

    def learn(self):
        self._result.learn(self.options._internal, self._input_variable_name)
        self._code_string = self._result.code
        self._code_snippet = _code.CodeSnippet(self._code_string)

    def __repr__(self):
        return """DetectTypesResult
===============
options = {}""".format(
            self.options
        )

    @classmethod
    def detect_data_types(cls, target, read_code, data, logger, input_variable_name):
        if target == "pyspark":
            # No data type detection possible when target==pyspark.
            return _code.CodeSnippet(read_code)
        detect_type_options = DetectTypesOptions(target=target)
        detect_type_options._internal.should_generate_code_with_input_from_read_apis = (
            True
        )

        detected_result = DetectTypesResultImpl(
            logger, data, detect_type_options, input_variable_name=input_variable_name
        )

        detect_types_code = detected_result.code
        if detect_types_code is not None and detect_types_code != "":
            read_code += os.linesep
            read_code += os.linesep
            read_code += detect_types_code
        return _code.CodeSnippet(read_code)
