# Copyright (c) Microsoft Corporation.  All rights reserved.

import pandas

from ._detect_types_result import _ColumnInfo
from . import _builder, _internals, _utils
from ._config import config

# See also: Python\CodeAccelerator\Internals\DomainTypeProfile*.cs


class DomainTypeProfileBuilderImpl(_builder.BuilderImpl):
    @classmethod
    def _default_target(cls):
        return config.get("default_target", "auto")

    @classmethod
    def _allowed_targets(cls):
        return [config.get("default_target", "auto")]

    _supported_domain_type_names = {
        "geographiccoordinate",
        "ipv6address",
        "ipv4address",
        "emailaddress",
        "usphonenumber",
        "uszipcode",
    }

    def __init__(self, logger, data):
        super().__init__()
        self._types_to_detect = list(self._supported_domain_type_names)

        type_of_data = type(data)
        self._input_data = data
        self._column_names_are_ints = False

        if type_of_data is pandas.DataFrame:
            self.column_infos = [_ColumnInfo(data[x], x, None) for x in data.columns]
            self._mode = "dataframe"
        elif isinstance(data, dict):
            self._mode = "dictionary"
            self.column_infos = [
                _ColumnInfo(data[column_name], column_name, None)
                for column_name in sorted(data.keys())
            ]
        else:
            try:
                data_as_list = list(data)
                self.column_infos = [_ColumnInfo(data_as_list, "Column_1", None)]
                self._mode = "list"
            except Exception:
                raise TypeError(
                    "The input data must be a pandas DataFrame, "
                    "a dictionary with string-valued lists as values "
                    "or a non-nested list of string values."
                )

        self._execution_data = data

        self._builder = _internals.DomainTypeProfileBuilder(
            logger, [x.native_object for x in self.column_infos]
        )

        self.target = self._default_target()

    @property
    def types_to_detect(self):
        return self._types_to_detect

    @types_to_detect.setter
    def types_to_detect(self, value):
        if value is None:
            return

        new_value = list(value)
        if not all(isinstance(x, str) for x in new_value):
            raise _utils.ProseException(
                "types_to_detect can only be assigned a list of string values."
            )
        for t in new_value:
            if t not in self._supported_domain_type_names:
                raise _utils.ProseException(
                    "Unknown type name '%s'. Supported type names are: %s"
                    % (t, ", ".join(sorted(self._supported_domain_type_names)))
                )
        self._types_to_detect = new_value

    def learn(self, logger):
        return DomainTypeProfileResultImpl(
            self._builder.learn(logger, self.types_to_detect)
        )


class DomainTypeProfileResultImpl(_builder.LearnResultImpl):
    def __init__(self, result: _internals.DomainTypeProfileResult):
        self._result = result

    def __getitem__(self, item) -> _internals.SingleColumnProfileResult:
        if isinstance(item, str):
            return self._result.get_profile_for_column(item)
        else:
            raise TypeError(
                "DomainTypeProfileResultImpl can only be indexed by string values."
            )

    def keys(self):
        return set(self._result.column_names)
