# Copyright (c) Microsoft Corporation.  All rights reserved.

from collections.abc import MutableMapping
from datetime import datetime, date, time
import pandas
import numpy


class _ExampleDict(MutableMapping):
    """A dictionary for storing examples that also accepts non-hashable keys.
    The given keys are converted into a tuple internally, if needed, so the users don't have to do it.
    Also, if the same input is mapped to a different output, a warning is emitted."""

    def __init__(self, logger, examples=None):
        """For internal use only."""
        self._logger = logger
        self._examples = {}
        if examples:
            for k, v in examples.items():
                self.__setitem__(k, v)

    def __getitem__(self, key):
        return self._examples[self._convert_to_tuple(key)]

    def __setitem__(self, key, val):
        self._examples[self._convert_to_tuple(key)] = val
        self._logger.track_event(
            [("SessionType", "DeriveValueBuilder"), ("InputForm", type(key).__name__)]
        )

    def _convert_to_tuple(self, key):
        def supported_base_type(key):
            return (
                isinstance(key, int)
                or isinstance(key, numpy.integer)
                or isinstance(key, float)
                or isinstance(key, numpy.floating)
                or isinstance(key, str)
                or isinstance(key, datetime)
                or isinstance(key, date)
                or isinstance(key, time)
                or isinstance(key, numpy.datetime64)
            )

        def all_supported(enumerable):
            return all(supported_base_type(i) for i in enumerable)

        if supported_base_type(key):
            return (key,)
        elif isinstance(key, tuple) and all_supported(key):
            return key
        elif (
            isinstance(key, list) or isinstance(key, pandas.Series)
        ) and all_supported(key):
            return tuple(key)
        else:
            raise ValueError(
                "Input row specified in unsupported format. Expected int, numpy.integer, float, numpy.floating, str, tuple, list, or pandas.Series"
            )

    def __delitem__(self, key):
        del self._examples[self._convert_to_tuple(key)]

    def __iter__(self):
        return iter(self._examples.keys())

    def __len__(self):
        return len(self._examples)

    def __repr__(self):
        return self._examples.__repr__()
