# Copyright (c) Microsoft Corporation.  All rights reserved.

from . import _builder, exceptions, _functions, _internals, _utils
from ._config import config

# Also see Python\CodeAccelerator\Internals\FindPatterns*.cs


class FindPatternsBuilderImpl(_builder.BuilderImpl):
    def __init__(self, logger, data):
        super().__init__()
        self._data = data
        self._builder = _internals.FindPatternsBuilder(logger, data)
        self.target = self.__class__._default_target()
        self.in_same_cluster = []
        self.in_different_clusters = []
        self.optimize_for_long_constants = None
        self.include_outlier_patterns = False
        self._long_constant_threshold = 40
        self._default_tokens = [
            "Space",
            "BinDigit",
            "Digit",
            "Lower",
            "Upper",
            "HexDigit",
            "Alpha",
            "AlphaSpace",
            "AlphaDigit",
            "AlphaDigitSpace",
            "Base64",
            "DotDash",
            "Punctuation",
            "Symbol",
            "Any",
            "TitleWord",
        ]
        self.custom_regex_tokens = []

    @classmethod
    def _default_target(cls):
        return config.get("default_target", "pandas")

    @classmethod
    def _allowed_targets(cls):
        return ["auto", "pandas", "pyspark"]

    def learn(self, logger):
        regex_tokens = [
            # The score of -1.0 is chosen by default. In general, this means that custom tokens
            # are given preference over most default tokens.
            _internals.RegexTokenSpec(name, regex, -1.0)
            for (name, regex) in self.custom_regex_tokens
        ]
        if self.optimize_for_long_constants is not None:
            optimize_for_long_constants = self.optimize_for_long_constants
        else:
            optimize_for_long_constants = any(
                s is not None and len(s) > self._long_constant_threshold
                for s in self._data
            )

        result = self._builder.learn(
            logger,
            self.in_same_cluster,
            self.in_different_clusters,
            self._default_tokens,
            regex_tokens,
            optimize_for_long_constants,
            self.include_outlier_patterns,
        )
        if result.error is not None:
            raise exceptions.LearnError(result.error)
        return FindPatternsLearnResultImpl(self._data, result, self.target)

    def __repr__(self):
        data_repr = _utils.shortened_repr(self._data)
        return """FindPatternsBuilder
===================
num_inputs = {}
inputs = {}
target = {}
in_same_cluster = {}
in_different_clusters = {}
optimize_for_long_constants = {}
include_outlier_patterns = {}""".format(
            len(self._data),
            data_repr,
            self.target,
            self.in_same_cluster,
            self.in_different_clusters,
            self.optimize_for_long_constants,
            self.include_outlier_patterns,
        )


class FindPatternsLearnResultImpl(_builder.LearnResultImpl):
    def __init__(self, data, result, target):
        self._data = data
        self._result = result
        self._target = target
        self._code_cache = {}

    @property
    def regexes(self):
        return self._result.regexes

    def code(self, task="classify"):
        lower_task = task.lower()
        if lower_task not in ["check", "classify"]:
            raise ValueError(
                "Invalid value for argument 'task'. Valid values are 'check' and 'classify'."
            )

        target = self._target
        if (lower_task, target) in self._code_cache:
            return self._code_cache[(lower_task, target)]

        function_name = (
            "identify_pattern" if lower_task == "classify" else "check_input"
        )
        code_string = self._code(lower_task, target, function_name)
        if target == "auto":
            ret = _functions.OneArgFunction(function_name, code_string)
        else:
            ret = _functions.TwoArgFunction(lower_task, code_string)

        self._code_cache[(lower_task, target)] = ret
        return ret

    def _code(self, lower_task, target, function_name):
        return self._result.code(lower_task, target, function_name)

    @property
    def preview_data(self):
        code = _functions.OneArgFunction(
            "identify_pattern", self._code("classify", "auto", "identify_pattern")
        )
        if code is None:
            return None

        ret = dict()
        for point in self._data:
            key = code(point)
            if key in ret:
                ret[key].append(point)
            else:
                ret[key] = [point]
        return ret

    def __repr__(self):
        return """FindPatternsLearnResult
=======================

-------------
preview_data
-------------
{}""".format(
            _utils.shortened_dict_repr(self.preview_data)
        )
