import atexit
import subprocess
import os
import weakref
import struct
from enum import Enum

from .signal_handler_util import add_signal_handlers


class _ProcessState(Enum):
    RUNNING = 1
    TERMINATED = 2


class BackgroundProcess(object):
    # Cache to hold binary name
    _process_cache = weakref.WeakValueDictionary()
    # Initialize path to private dotnet executable
    try:
        from dotnetcore2 import runtime
        runtime.ensure_dependencies()
        _dotnet_path = runtime.get_runtime_path()
    except ValueError as e:
        if format(e) == 'Unsupported Linux distribution':
            raise Exception('.NET Core is not installed properly. Note that Linux distros other than Ubuntu are not '
                            'supported.')
        raise e

    @classmethod
    def cache_cleanup(cls):
        for bg_object in cls._process_cache.values():
            bg_object._dispose_process()

    def __new__(cls, binary_name, guid, process_terminator):
        """
        Control the creation of BackgroundProcess objects based on :binary_name and :guid. When the process
        has to be terminated, it invokes process_terminator handler to inform dotnet process to shutdown.
        :param binary_name:
        :param guid:
        :param process_terminator: A Python callable method that takes BackgroundProcess object as parameter to communicate
        terminate message to dotnet process.
        :return:
        """

        if not callable(process_terminator):
            raise ValueError('Parameter process_terminator should be callable.')
        key = binary_name + '_' + guid + '_' + str(id(process_terminator))
        # Hold the reference to the instance, if exist.
        instance = cls._process_cache.get(key, None)
        # Although instance might be in the cache, but it might have get collected.
        # So make sure we have a dotnet process running for this instance.
        if instance is None or instance._process_state != _ProcessState.RUNNING:
            instance = super(BackgroundProcess, cls).__new__(cls)
            base_path = os.path.abspath(os.path.dirname(__file__))
            binary_path = os.path.join(base_path, binary_name)
            if not os.path.exists(binary_path):
                raise ChildProcessError("{} doesn't exist.".format(binary_path))

            dotnet_process = subprocess.Popen([cls._dotnet_path, binary_path, guid], stdin=subprocess.PIPE,
                                              stdout=subprocess.PIPE, bufsize=0)

            # read dotnet process startup status
            length = int.from_bytes(dotnet_process.stdout.read(4), byteorder='little', signed=True)
            received_message = dotnet_process.stdout.read(length).decode("ascii")
            if received_message != 'success':
                raise RuntimeError('Failed to start "dotnet {}" : {}'.format(binary_name, received_message))

            # save instance in cache
            instance._dotnet_process = dotnet_process
            cls._process_cache[key] = instance
        return instance

    def __init__(self, binary_name, guid, process_terminator):
        """
        Initializes the fields for :binary_name and :guid
        :param binary_name:
        :param guid:
        :param process_terminator:
        """

        # Ensure either we haven't set these attributes or we are resetting to same value in case this was already
        # created object.
        assert not hasattr(self, '_binary_name') or self._binary_name == binary_name
        assert not hasattr(self, '_guid') or self._binary_name == binary_name
        assert not hasattr(self, '_process_terminator') or self._process_terminator == process_terminator

        self._binary_name = binary_name
        self._guid = guid
        self._process_terminator = process_terminator
        self._process_state = _ProcessState.RUNNING

    def __del__(self):
        # if this was the last object that referred the process, then kill the process as well
        # only if it was started in first place.
        if hasattr(self, '_dotnet_process'):
            self._dispose_process()

    def communicate(self, message_to_send):
        """
        Sends message to underlying dotnet process and receives back the reply.
        The protocol is to send message in 2 phases:
        - Send the message length (4 bytes representing signed int) in little-endian order
        - Send the actual message.
        Likewise, the message is received in 2 phases:
        - Message length (4 bytes representing signed int) in little-endian order
        - The actual message.
        :param message_to_send:
        :return:
        """

        # If process already exited, no communication possible
        if self._process_state != _ProcessState.RUNNING:
            return None

        try:
            message = message_to_send.SerializeToString()
            # pack the message length in 4 bytes that represent signed int and is in little-endian order.
            self._dotnet_process.stdin.write(struct.pack('<I', len(message)))
            self._dotnet_process.stdin.write(message)
            # unpack the message length of 4 bytes that represent signed int and is in little-endian order.
            bytes_to_read = int.from_bytes(self._dotnet_process.stdout.read(4), byteorder='little', signed=True)

            received_message = b''
            total_bytes_read = 0
            bytes_read_per_iter = -1
            bytes_remaining = bytes_to_read
            while bytes_remaining != 0 and bytes_read_per_iter != 0:
                # read partial result
                partial_received = self._dotnet_process.stdout.read(bytes_to_read)
                received_message += partial_received

                # update counters
                bytes_read_per_iter = len(partial_received)
                total_bytes_read = total_bytes_read + bytes_read_per_iter
                bytes_remaining = bytes_remaining - bytes_read_per_iter

            if bytes_read_per_iter == 0:
                raise RuntimeError('Nothing to read from stream.')
            if total_bytes_read != bytes_to_read:
                raise RuntimeError(
                    'Expected to read {} bytes, but read {} bytes.'.format(bytes_remaining, total_bytes_read))
            return received_message
        except ValueError as e:
            # I/O operation failure is because dotnet process died.
            if format(e) == 'I/O operation on closed file':
                self._update_process_state_if_not_alive()
                return
            raise RuntimeError('Internal error has occurred.')
        except OSError:
            self._update_process_state_if_not_alive()

    def _update_process_state_if_not_alive(self):
        """ Polls for 0.1 seconds to see if process is alive and if not updates the state to TERMINATED."""
        if not self._poll_process_alive(0.1):
            self._process_state = _ProcessState.TERMINATED
        return

    def _dispose_process(self):
        """
        Terminates the underlying dotnet process. First it attempts to terminate gracefully,
        but if that doesn't work, will forcefully kill the process.
        :return:
        """
        if self._process_state != _ProcessState.RUNNING:
            return

        # Try terminating gracefully
        pid = self._dotnet_process.pid
        result = self._process_terminator(self)
        if result == 'success':
            # If process still alive, kill it
            if self._poll_process_alive():
                self._kill_process()
        else:
            self._kill_process()
        self._process_state = _ProcessState.TERMINATED

        # if process was killed, make sure it was killed for real
        if self._poll_process_alive():
            raise OSError("Couldn't kill process {}".format(pid))

    def _poll_process_alive(self, timeout=5):
        """
        Polls if process is alive for :timeout seconds.
        :param timeout:
        :return: Returns True if process is alive, else False
        """
        try:
            self._dotnet_process.wait(timeout)
            # Process no longer present. Return
            return False
        except subprocess.TimeoutExpired:
            # Process still alive
            pass
        return True

    def _kill_process(self):
        """
        Forcefully kill the process
        :return:
        """
        try:
            self._dotnet_process.kill()
        except ProcessLookupError:
            # process might have been killed already.
            pass


# Register exit handler
atexit.register(BackgroundProcess.cache_cleanup)
# Register signal handlers
add_signal_handlers()
