import weakref


class CacheManager(object):
    """
    This is analogous to C# CacheManager. The objects present in Python's cache manager assume that their
    counter part objects are present in C#'s cache manager. When Python objects are collected, entries are
    automatically removed from Python side as well as from C# side.
    """

    # Holds cache managers. Each instance of CacheManager corresponds to a
    # unique dotnet process
    _cache_managers = weakref.WeakValueDictionary()

    def __new__(cls, background_process_id):
        if background_process_id not in cls._cache_managers:
            instance = super(CacheManager, cls).__new__(cls)
            cls._cache_managers[background_process_id] = instance
        else:
            instance = cls._cache_managers[background_process_id]
        assert instance is not None
        return instance

    def __init__(self, background_process_id):
        # Control is transferred from __new__ regardless if instance is created or not
        # This makes sure we don't override the already created cache
        if hasattr(self, '_objects_cache'):
            return

        # dictionary to cache objects
        self._objects_cache = weakref.WeakValueDictionary()

        # dictionary to cache fetcher objects. 
        # Objects and fetchers use different id calculations and can collide,
        # so they are stored in separate caches.
        self._fetcher_objects_cache = weakref.WeakValueDictionary()

        self._background_process_id = background_process_id

    def add_object(self, reference_id, object_to_cache):
        """
        Adds the object to the cache.
        Ensures that the cache contains the passed in :object_to_cache indexed by the specified :reference_id.
        This method is *almost always* called by __init__ of objects.
        :param reference_id:
        :param object_to_cache:
        :return:
        """
        cached_object = self._objects_cache.get(reference_id, None)
        if cached_object is not None:
            assert cached_object is object_to_cache
            return
        self._objects_cache[reference_id] = object_to_cache

    def add_fetcher_object(self, fetcher_object_to_cache):
        """
        Adds the fetcher object to the cache and returns the id generated by calling python's id()
        on the argument.
        If the object is already present, just returns the id.
        :param fetcher_object_to_cache:
        :return: fetcher_id
        """
        id_for_object_to_cache = id(fetcher_object_to_cache)
        # If we cached a different object in the past that had same id, it might have got collected
        # and hence its entry removed from fetcher cache. This is a new object and hence will get cached.
        if id_for_object_to_cache not in self._fetcher_objects_cache:
            self._fetcher_objects_cache[id_for_object_to_cache] = fetcher_object_to_cache
        
        return id_for_object_to_cache

    def get_object(self, reference_id):
        """
        Gets the cached object. If reference_id doesn't exist in cache, it means
        that the object was created on C# side. Just return None so caller can create
        a fresh instance and cache it.
        This method is *almost always* called in methods that gets object's referenceId as
        return value from C# side.
        :param reference_id:
        :return:
        """

        return self._objects_cache.get(reference_id, None)

    def get_fetcher_object(self, reference_id):
        """
        Gets the cached fetcher object.
        :param reference_id:
        :return:
        """

        return self._fetcher_objects_cache.get(reference_id, None)
        

    def get_object_cache_count(self):
        """
        Returns no. of entries present in this cache manager
        :return:
        """
        return len(self._objects_cache)

    def get_fetcher_object_cache_count(self):
        """
        Returns no. of fetcher object entries present in this cache manager
        :return:
        """
        return len(self._fetcher_objects_cache)
