﻿import signal

import sys


def _terminate_handler(signum=None, frame=None):
    # Enables calling functions registered using `atexit.register()`
    sys.exit(signum)


def _add_signal_handler(signalnum):
    """
    Adds handler to the signalnum. If there was old signal handler, then it will first
    execute handler followed by old handler
    :param signalnum:
    :param new_handler:
    :return:
    """

    old_handler = signal.getsignal(signalnum)

    if old_handler is signal.SIG_IGN:
        return

    if old_handler is signal.SIG_DFL:
        # Only overwrite signal handler for SIGTERM and SIGINT signals

        if signalnum in _terminate_signals:
            signal_handler_wrapper = _terminate_handler

        # For other signals, we want to generate core-dump, so we won't overwrite.
        else:
            signal_handler_wrapper = old_handler
    else:
        # Add wrapper if old_handler were anyway overwritten
        def signal_handler_wrapper(signum=None, frame=None):
            """
            Execute old_handler() followed by terminate_handler()
            """
            try:
                if callable(old_handler):
                    old_handler(signum, frame)
            finally:
                _terminate_handler(signum, frame)

    signal.signal(signalnum, signal_handler_wrapper)


_terminate_signals = [signal.SIGTERM, signal.SIGINT]
if sys.platform == 'win32':
    _terminate_signals.append(signal.SIGBREAK)
else:
    _terminate_signals.append(signal.SIGHUP)


def add_signal_handlers():
    """
    http://man7.org/linux/man-pages/man7/signal.7.html
    Adds signal handlers to terminate gracefully.

    Here are the actions performed for signals based on default actions of those signals.

    1. Term: These are the signals whose default action is to terminate the process. For such signals,
    if their current signal handler is the default signal handler i.e. signal.SIG_DFL, this method overwrites it with
    terminate handler that just calls sys.exit(). If there is a non-default handler attached. it will execute it
    first before executing the terminate handler.
    E.g. SIGTERM, SIGINT, SIGHUP, SIGBREAK.

    2. Core: These are the signals whose default action is terminate the process and dump core. For such signals,
    if their current signal signal handler is the default signal handler i.e. signal.SIG_DFL, retains that default
    signal handler, the reason being we want to continue getting core dump. However if the signal handler for such
    signals is already overwritten, then overwrite once again with terminate handler such that the previous signal
    handler is called before the terminate handler (just as Term case).
    E.g. SIGQUIT, SIGILL, SIGABRT, SIGFPE, SIGSEGV

    3. Ign: For all the signals which has default action set to signal.SIG_IGN, this method will continue ignoring the
    signal and won't register terminate handler.

    4. Kill: signal.SIGKILL and signal.SIGSTOP are the 2 signals that cannot be caught, blocked or ignored. This
    method continues to honor that.

    :return:
    """
    # Signals that just terminate
    _add_signal_handler(signal.SIGTERM)
    _add_signal_handler(signal.SIGINT)

    # Signals that terminates and dump core.
    if sys.platform == 'win32':
        _add_signal_handler(signal.SIGBREAK)
    else:
        _add_signal_handler(signal.SIGHUP)
        _add_signal_handler(signal.SIGQUIT)
    _add_signal_handler(signal.SIGILL)
    _add_signal_handler(signal.SIGABRT)
    _add_signal_handler(signal.SIGFPE)
    _add_signal_handler(signal.SIGSEGV)
