# Copyright (c) Microsoft Corporation.  All rights reserved.

import itertools


class _IterableStreamer(object):
    """A class that wraps an iterable object for streaming across XplatPyNet."""

    def __init__(self, data_to_iterate):
        """Create an iterable streamer for XplatPyNet."""
        # noinspection PyBroadException
        try:
            self._length = len(data_to_iterate)
        except Exception:
            self._length = None

        self._data_to_iterate = data_to_iterate
        self._iterator = iter(self._data_to_iterate)
        self._current_element = None
        self._end_of_stream = False

    def move_next(self):
        """Returns true iff the stream has more elements."""
        if self._end_of_stream:
            return False
        try:
            self._current_element = next(self._iterator)
        except StopIteration:
            self._current_element = None
            self._end_of_stream = True
            return False
        return True

    def get_current(self):
        """Retrieves the current element in the stream."""
        return self._current_element

    def get_length(self):
        """Retrieves the length of the stream if possible."""
        return self._length

    def bulk_read(self, n):
        """Reads the next n elements of the stream."""
        if n is None:
            return list(self._iterator)
        else:
            result = []
            for i in range(n):
                while self.move_next():
                    result.append(self.get_current())
            return result

    def reset(self):
        """Reset the iterator and other elements data."""
        self._iterator = iter(self._data_to_iterate)
        self._current_element = None
        self._end_of_stream = False
