# Copyright (c) Microsoft Corporation.  All rights reserved.

import math
from io import SEEK_SET, SEEK_END, SEEK_CUR

_READ_TILL_EOF = -1
_NODATA_VAL = b""


def _read_from_stream(stream, bytes_to_read):
    """ Read bytes from raw stream """
    assert (
        bytes_to_read is None or bytes_to_read >= -1
    ), "_read_bytes's size should be None or value >= -1"
    total_bytes_read = 0
    data = []
    bytes_remaining = bytes_to_read
    while bytes_remaining == _READ_TILL_EOF or bytes_remaining != 0:
        chunk = stream.read(bytes_remaining)
        bytes_read_curr_iter = len(chunk)
        if bytes_read_curr_iter == 0:
            break
        data.append(chunk)
        total_bytes_read += bytes_read_curr_iter
        if bytes_to_read == _READ_TILL_EOF:
            bytes_remaining = _READ_TILL_EOF
        else:
            bytes_remaining -= bytes_read_curr_iter
    return b"".join(data)


class ProseReaderForNonSeekableStream(object):
    """
    Prose non-seekable stream reader : A buffered stream reader that buffers the contents read from
    a non-seekable stream.
    """

    """
     Imagine a _DEFAULT_BLOCK_SIZE of N bytes, the buffer is a list of list. Every element of
     outer list stores a list containing precisely _DEFAULT_BLOCK_SIZE bytes.
     Below figure shows x blocks of data containing between (x * N) and (x + 1 * N) bytes.
     The last block can have fewer than N bytes.

               ----------------------------
     BlockID 0 |0,.................,.,N-1 |
               ----------------------------
     BlockID 1 |N,..................,2N-1 |
               ----------------------------
     BlockID 2 |2N,...................... |
               ----------------------------
     BlockID x |xN,...................... |
               ----------------------------

     The streaming is done inside blocks of size _DEFAULT_BLOCK_SIZE bytes. It gives ability to seek backward
     and forward on buffer despite the underlying stream being non-seekable.

     Purpose : For PythonSDK file ingestion, stream will be passed to the APIs from which data
     needs to be read. The PythonSDK APIs needs a buffer reader that can start reading from the
     position a non-seekable stream came up with and buffer the contents in memory. The buffer reader should
     provide a virtual view of the underlying stream with its own buffer pointer that can go forward
     and backward on data present in buffer.

     Here is the list of functionality that a non-seekable stream reader should support for PythonSDK
     and that is difficult to accomplish using Python's io library's BufferedReader:
     1.  It should start reading the contents from original position and buffer them in memory.
     2.  When asked to seek to to an arbitrary position, it should reset the reader pointer
         to that position despite the underlying stream is non-seekable. This ability is
         needed during multiple learn() calls when contents from stream has to be re-read. Also
         during preview_data, the contents need to be passed to the generated code from the beginning.
     3.  Reader's current position should always be relative to the original position from where it
         started buffering.

     ProseReaderForNonSeekableStream provides following APIs :
         1.  tell()  : Return the buffer position regardless of current position of underlying
             stream.

         2.  seek()  : Depending on value of whence, it will perform the following:
             a.  SEEK_SET : The buffer position will be set to the passed offset.
             b.  SEEK_CUR : The buffer position will be moved by offset bytes (either forward
                 or backward).
             c.  SEEK_END : Since we don't want to seek to the end of stream, because the stream
                 can be long, an error will be thrown.

         3.  read()  : Chunks of BLOCK_SIZE will be read from stream, saved inside buffer
             and returned the desired count of bytes starting from buffer position. If data requested
             is already present, it will be just read from buffer and returned back.
     """

    _DEFAULT_BLOCK_SIZE = 4096

    def __init__(self, non_seekable_stream, block_size=_DEFAULT_BLOCK_SIZE):
        """ Creates an instance of ProseReaderForNonSeekableStream """
        if non_seekable_stream.seekable():
            raise ValueError("Only use with non-seekable stream")

        self._raw_stream = non_seekable_stream
        self._raw_pos = 0
        self._block_size = block_size
        self._buffer = []
        self._buffer_size = 0

        # self._buf_pos is the current position of buffer pointer.
        # read() method will read from (and including) buf_pos onwards.
        # seek() method will move the buf_pos.
        self._buf_pos = 0

    def __repr__(self):
        contents = [
            "seekable = False",
            "_raw_pos: {}".format(self._raw_pos),
            "_block_size: {}".format(self._block_size),
            "_buffer_size: {}".format(self._buffer_size),
            "_buf_pos: {}".format(self._buf_pos),
        ]
        return "\n".join(contents)

    def _get_block_id_of_pos(self, position):
        """
        Returns block id that contains the given :position:
        :param position: Position for which block id should be found.
        :return: Block id
        """
        return math.floor(position / self._block_size)

    def _verify_buffer_integrity(self, expected_buffer_size=None):
        """
        Test-method to ensure that all blocks but last one are BLOCK_SIZE bytes long.
        Also verifies if buffer size matches expected_buffer_size to ensure that
        only data that is needed is saved inside the buffer.
        Raises BufferError if not.
        """
        buffer_len = len(self._buffer)
        if buffer_len == 0:
            return

        for block_id in range(0, buffer_len - 1):
            chunk = self._buffer[block_id]
            chunk_len = len(chunk)
            if chunk_len != self._block_size:
                raise BufferError(
                    "{} bytes present at block_id: {} / {}".format(
                        chunk_len, block_id, buffer_len - 1
                    )
                )
        last_chunk_len = len(self._buffer[buffer_len - 1])
        if last_chunk_len > self._block_size:
            raise BufferError(
                "{} bytes present at last block_id: {} / {}".format(
                    last_chunk_len, buffer_len - 1, buffer_len - 1
                )
            )

        if (
            expected_buffer_size is not None
            and self._buffer_size != expected_buffer_size
        ):
            raise BufferError(
                "Buffer size mismatch. Expected: {}, Actual: {}".format(
                    expected_buffer_size, self._buffer_size
                )
            )

    def _read_from_stream_into_buffer(self, block_id_to_fill_up_to=_READ_TILL_EOF):
        """
        Reads the data from raw stream and start appending it into the buffer.
        The data read from stream is appended with granularity of block_size against
        the newer block_ids added at the end of existing buffer. The last block_id to
        get filled is :block_id_to_fill_up_to:.

        If :block_id_to_fill_up_to: is -1, then reads entire contents starting from raw_pos.
        This is the only method that updates raw_pos.
        """
        # If block is already in buffer, just return
        if 0 <= block_id_to_fill_up_to < len(self._buffer):
            return

        if block_id_to_fill_up_to == _READ_TILL_EOF:
            size_to_read = _READ_TILL_EOF
        else:
            size_to_read = (
                block_id_to_fill_up_to - len(self._buffer) + 1
            ) * self._block_size
        all_contents = _read_from_stream(self._raw_stream, size_to_read)
        contents_size_read = len(all_contents)

        # If nothing was read, return
        if contents_size_read == 0:
            return
        self._raw_pos += contents_size_read

        start_pos = 0

        # Put all_contents in the buffer
        while True:
            end_pos = start_pos + self._block_size
            # Read data of block_size and save in buffer
            chunk = all_contents[start_pos:end_pos]
            chunk_len = len(chunk)
            if chunk_len == 0:
                break
            self._buffer.append(chunk)
            self._buffer_size += chunk_len
            if chunk_len < self._block_size:
                break
            start_pos = end_pos

    def _read_from_buffer(self, size):
        """
        Should only be called after checks are made that data exist in buffer.
        Read all the chunks from block_ids [start_block_id to end_block_id) and returns
        the data of :size: out of it starting from buffer index buf_pos.
        If size == -1 or None, entire contents of buffer starting from buf_pos will be read.
        Advance the buf_pos by no. of bytes read.
        :return:
        """
        chunks = []
        # Get block id containing buf_pos so read can happen starting from that block.
        start_block_id = self._get_block_id_of_pos(self._buf_pos)

        if size == _READ_TILL_EOF:
            end_block_id = len(self._buffer)
        else:
            # Get the last block id to read to satisfy the read() request of `size`.
            # Since read starts from buf_pos, need to read (size - 1) more bytes.
            last_block_id_to_read = self._get_block_id_of_pos(
                self._buf_pos + (size - 1)
            )
            end_block_id = min(last_block_id_to_read + 1, len(self._buffer))

        for block_id in range(start_block_id, end_block_id):
            chunks.append(self._buffer[block_id])

        all_chunks = b"".join(chunks)

        start_pos = self._buf_pos - (start_block_id * self._block_size)

        if size == _READ_TILL_EOF:
            end_pos = len(all_chunks)
        else:
            end_pos = start_pos + size

        out = all_chunks[start_pos:end_pos]
        num_of_bytes_read = len(out)
        if size != _READ_TILL_EOF and size < num_of_bytes_read:
            raise ValueError(
                "Expected to read {} bytes, but read {} bytes.".format(
                    size, num_of_bytes_read
                )
            )

        # Ensure we have not read more data from buffer than needed.
        assert (
            start_pos < self._block_size
        ), "Read too much prefix data from buffer into chunks"
        assert (
            len(all_chunks) - end_pos < self._block_size
        ), "Read too much suffix data from buffer into chunks"

        # Update buf_pos only by amount buffer was read
        self._buf_pos += num_of_bytes_read
        return out

    @property
    def raw_stream(self):
        return self._raw_stream

    @staticmethod
    def seekable():
        return False

    def seek(self, offset, whence=SEEK_SET):
        """
        Seek the buffer position to the offset depending on the value
        of whence.
        - SEEK_SET : Set buf_pos to offset
        - SEEK_CUR : Set buf_pos to (buf_pos + offset)
        :param offset: offset to seek to
        :param whence: Relative position from where to seek.
        :return:
        """

        if whence == SEEK_SET:
            if offset < 0:
                raise ValueError("negative seek position {}".format(offset))
            self._buf_pos = offset
            return self._buf_pos
        elif whence == SEEK_CUR:
            new_pos = self._buf_pos + offset
            if new_pos < 0:
                raise ValueError(
                    "negative seek position {} with respect to position: {}".format(
                        offset, self._buf_pos
                    )
                )
            self._buf_pos = new_pos
            return self._buf_pos
        elif whence == SEEK_END:
            raise ValueError(
                "Seeking to the end of stream is not supported because it can be "
                "problematic for larger streams."
            )
        else:
            raise ValueError("whence value {} unsupported".format(whence))

    def tell(self):
        return self._buf_pos

    def read(self, size=None):

        if size == 0:
            return _NODATA_VAL

        if size is None or size < -1:
            size = _READ_TILL_EOF

        # Find the block_id where end position lies.
        # If read till end (size == -1), then block_id_to_read_up_to is -1 which is read till EOF
        block_id_to_fill_up_to = (
            self._get_block_id_of_pos(self._buf_pos + (size - 1)) if size > 0 else -1
        )

        # If block_id_to_read_up_to is already present in buffer,
        # below will be no-op
        self._read_from_stream_into_buffer(block_id_to_fill_up_to)

        return self._read_from_buffer(size)


class ProseReaderForSeekableStream(object):
    """
    Prose seekable stream reader : A stream reader that remembers the position of underlying
    stream when it was passed to the reader and performs read(), tell() and seek() operations
    with respect to that position.
    """

    """
    Purpose : For PythonSDK file ingestion, stream will be passed to the APIs from which data
    needs to be read. The incoming stream could have been seeked/read till position > 0.
    The PythonSDK APIs needs a stream reader that can start reading from the position the stream
    came up with. The stream reader should provide a virtual view of contents of underlying stream
    starting from original position.

    Here is the list of functionality that a seekable stream reader should support for PythonSDK
    and that is difficult to accomplish using just the Python's io library's BufferedReader:
    1.  The reader should remember the original position of the underlying stream when it was
        passed to the reader.
    2.  It should start reading the contents from original position.
    3.  When asked to seek to to an arbitrary position, it should reset the reader pointer
        to a position relative to the original position of underlying stream. This ability is
        needed during multiple learn() calls when contents from stream has to be re-read. Also
        during preview_data, the contents need to be passed to the generated code from the beginning.
    4.  Reader's current position should always be relative to the original position.

    Stream types :
        1.  Seekable :  For streams that are seekable, all the operations on the reader will directly
            happen on the underlying stream without buffering.
        2.  Non-seekable :  For non-seekable streams,  the reader will remember the original
            position to the stream. During read() operation, the reader will start reading from the
            original position and buffer the contents in blocks as described above. Operations like
            tell() and seek() will also be done with respect to the original position of the stream.

    ProseReaderForSeekableStream provides following APIs :
        1.  tell()  : Returns the current position of stream relative to the original position.

        2.  seek()  : Depending on value of whence, it will perform the following:
            a.  SEEK_SET : It will set the current position of stream to original position + offset.
            b.  SEEL_CUR : It will add the current position to the passed offset. If offset was < 0
                and the new position takes it to before original position it throws the error.
            c.  SEEK_END : Since we don't want to seek to the end of stream, because the stream
                can be long, an error will be thrown.

        3.  read()  : Returns the data read directly from underlying stream without
                buffering starting from current position of the stream.
    """

    def __init__(self, seekable_raw_stream):
        """ Creates an instance of ProseReaderForSeekableStream """

        if not (
            hasattr(seekable_raw_stream, "seekable"),
            hasattr(seekable_raw_stream, "read"),
            hasattr(seekable_raw_stream, "seek"),
            hasattr(seekable_raw_stream, "tell"),
        ):
            raise ValueError(
                "seekable_raw_stream should have attributes : seekable, read, seek and tell."
            )
        if not seekable_raw_stream.seekable():
            raise ValueError("Only use with seekable stream")
        self._raw_stream = seekable_raw_stream
        self._original_raw_pos = self._raw_tell()

    def _raw_seek(self, offset, whence):
        """ Seek on raw_stream """
        return self._raw_stream.seek(offset, whence)

    def _raw_tell(self):
        """ Tell on raw_stream """
        return self._raw_stream.tell()

    @property
    def raw_stream(self):
        return self._raw_stream

    @staticmethod
    def seekable():
        return True

    def seek(self, offset, whence=SEEK_SET):
        """
        Seek the buffer position to the offset depending on the value
        of whence.
        - SEEK_SET : Seek to (original_raw_pos + offset)
        - SEEK_CUR : Seek to (current_position + offset)
        :param offset: offset to seek to
        :param whence: Relative position from where to seek.
        :return: Returns the new position.
        """

        if whence == SEEK_SET:
            new_pos = self._original_raw_pos + offset
            if new_pos < self._original_raw_pos:
                raise ValueError("negative seek position {}".format(offset))
            return self._raw_seek(new_pos, whence)
        elif whence == SEEK_CUR:
            new_pos = self._raw_tell() + offset
            if new_pos < self._original_raw_pos:
                raise ValueError(
                    "negative seek position {} with respect to position: {}".format(
                        offset, self._original_raw_pos
                    )
                )
            return self._raw_seek(offset, whence)
        elif whence == SEEK_END:
            raise ValueError(
                "Seeking to the end of stream is not supported because it can be "
                "problematic for larger streams."
            )
        else:
            raise ValueError("whence value {} unsupported".format(whence))

    def tell(self):
        """
        Returns the current position of buffer position
        :return:
        """
        pos = self._raw_tell() - self._original_raw_pos
        assert (
            pos >= 0
        ), "Underlying stream position is at {} which is before the original position {}".format(
            self._raw_tell(), self._original_raw_pos
        )
        return pos

    def read(self, size=None):
        """
        Reads the contents from raw_stream.
        :param size: Size of contents to read
        :return: Data read.
        """

        if size == 0:
            return _NODATA_VAL

        if size is None or size < -1:
            size = _READ_TILL_EOF

        return _read_from_stream(self._raw_stream, size)
