# Copyright (c) Microsoft Corporation.  All rights reserved.

import os
import pathlib
from . import _code
from ._base_telemetry import BaseTelemetry
from ._detect_types_result import DetectTypesResultImpl
from ._api._telemetry_utils import _class_properties
from ._api._read_parquet_options import ReadParquetOptions

RESULT_VARIABLE_NAME = "df"


def _get_pyspark_code(filename, columns):
    code_string = """from pyspark.sql import SparkSession
from pyspark.sql.functions import col

spark = SparkSession.builder.master('local').getOrCreate()
{} = spark.read.parquet({})""".format(
        RESULT_VARIABLE_NAME, filename
    )

    if columns:
        code_string += os.linesep
        code_string += "{} = {}.select({})".format(
            RESULT_VARIABLE_NAME, RESULT_VARIABLE_NAME, columns
        )

    code_string += os.linesep
    code_string += RESULT_VARIABLE_NAME

    return code_string


def _get_pandas_code(filename, columns):
    code_string = """import pandas
{} = pandas.read_parquet({}, columns={})
{}""".format(
        RESULT_VARIABLE_NAME, filename, columns, RESULT_VARIABLE_NAME
    )

    return code_string


class ReadParquetResultImpl(BaseTelemetry):
    @property
    def _internal(self):
        return self

    @property
    def python_logger(self):
        return self._python_logger

    def __init__(self, logger, filename, options):
        # TODO: Not needed if file_name was a variable instead of path.
        self._python_logger = logger
        self._filename = 'r"{}"'.format(str(pathlib.Path(filename)))
        self._options = ReadParquetOptions() if options is None else options
        self._code_snippet = None
        self._code_string = None
        self.learn()

    @property
    def code(self):
        return self._code_string

    @property
    def options(self):
        return self._options

    @property
    def data(self):
        if self._code_snippet is None:
            return None

        return self._code_snippet.evaluate()

    def learn(self):
        target = self._options.target
        columns = self._options.columns

        try:
            columns = columns if isinstance(columns, list) else None
            if target == "pandas":
                self._code_string = _get_pandas_code(self._filename, columns)
            elif target == "pyspark":
                self._code_string = _get_pyspark_code(self._filename, columns)
            else:
                raise ValueError(
                    "Invalid value '{}' for 'target'. Expected a value from ['pandas', 'pyspark'].".format(
                        target
                    )
                )

            self._code_snippet = _code.CodeSnippet(self._code_string)

            update_columns = True if not columns else False
            detect_datatypes = self._options.detect_data_types

            if update_columns or detect_datatypes:
                # Evaluate data only if needed
                data = self.data

                if update_columns and hasattr(data, "columns"):
                    self._options.columns = data.columns

                if detect_datatypes:
                    self._code_snippet = DetectTypesResultImpl.detect_data_types(
                        target,
                        self._code_string,
                        data,
                        self._internal.python_logger,
                        RESULT_VARIABLE_NAME,
                    )
                    self._code_string = repr(self._code_snippet)
        finally:
            # This will create 1 extra telemetry entry for update in addition
            # to what is logged from .utils. However this will capture the extra
            # information like target and columns.
            columns_count = (
                len(self._options.columns) if self._options.columns is not None else 0
            )
            properties = _class_properties("learn", "ReadParquetResult")
            properties.append(("PythonTarget", target))
            properties.append(("ColumnsCount", str(columns_count)))
            self._python_logger.track_event(properties)

    def __repr__(self):
        return """ReadParquetResult
===============
filename = {}
options = {}""".format(
            self._filename, self._options
        )
